"""
functions for handling the MERtools .sel format
"""
import warnings
from collections.abc import Mapping
from pathlib import Path
from typing import Union

import numpy as np
import scipy.io
from astropy.io import fits

from marslab.compat.mertools import (
    MERSPECT_MSL_COLOR_MAPPINGS,
    MERSPECT_M20_COLOR_MAPPINGS,
    MERSPECT_COLOR_MAPPINGS,
)
from marslab.imgops.regions import select_roi_by_ix, make_roi_hdu


def is_sel_file(roi_path: Union[str, Path]) -> bool:
    """
    is this file a MERSpect .sel file?
    """
    # TODO: scipy.io.readsav does not close the buffer correctly
    #  on finding an invalid signature.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", ResourceWarning)
        try:
            sel = scipy.io.readsav(roi_path)
            assert set(sel.keys()).issubset(
                {
                    "erasecolor",
                    "lseltemp",
                    "rseltemp",
                    "region_info",
                    "left_pos",
                    "right_pos",
                    "sel_file_format_major_version",
                    "sel_file_format_minor_version",
                    "sel_file_format_date",
                }
            )
            return True
        except AssertionError:
            # this is apparently an IDL .sav file, but not a
            # properly-formatted .sel file
            return False
        # scipy.io.readsav uses general Exception
        except Exception as exception:
            if "Invalid SIGNATURE" in str(exception):
                # this is not an IDL .sav file but rather something else
                return False
            # presumably FileNotFoundError, you passed an integer, stuff
            # like that
            raise


def roi_color_ix_to_color_name(color_ix: int, instrument: str = "MCAM") -> str:
    """
    convert MERSpect ROI color index to color name
    """
    # this works because colors are listed in the dictionaries in
    # marslab.compat.mertools in the order
    # they are listed in the MERSpect IDL code. No explicit mapping
    # between MERSpect ROI color indices and actual colors exists.
    if instrument in ["MCAM", "PCAM"]:
        return list(MERSPECT_MSL_COLOR_MAPPINGS.keys())[color_ix]
    elif instrument == "ZCAM":
        # The unit offset to the color indices is because "0" is reserved
        # in MERspect as something like "blank canvas." The "erase"
        # capability has a different color associated with it.
        return list(MERSPECT_M20_COLOR_MAPPINGS.keys())[color_ix - 1]
    else:
        raise ValueError(
            "I don't have ROI color information about "
            + instrument
            + ". I know about PCAM, MCAM, and ZCAM."
        )


def roi_hdu_from_sel(
    sel_image: np.ndarray,
    color_ix: int,
    eye_name: str,
    instrument: str = "MCAM",
    sel_fn: str = None,
) -> fits.PrimaryHDU:
    """
    make an individual HDU for a marslab roi FITS file
    from an array generated from a MERSpect .sel file
    """
    roi_array = select_roi_by_ix(sel_image, color_ix)
    color_name = roi_color_ix_to_color_name(color_ix, instrument)
    merspect_metadata = {"EYE": eye_name, "SOURCEFN": Path(sel_fn).name}
    roi_hdu = make_roi_hdu(roi_array, color_name, merspect_metadata)
    roi_hdu.name = color_name + " " + eye_name
    return roi_hdu


def sel_to_roi(sel_fn: Union[str, Path], instrument: str) -> fits.HDUList:
    """
    convert a MERSpect .sel file to a marslab roi file -- a FITS file with a
    single HDU per eye per ROI
    """
    sel = scipy.io.readsav(sel_fn)
    sel_images = {}
    # add each eye if present
    for eye_name, sel_key in (("left", "lseltemp"), ("right", "rseltemp")):
        if isinstance(sel[sel_key], np.ndarray):
            # IDL y-axis indexing is flipped compared to numpy, so flip
            # images here
            sel_images[eye_name] = np.flipud(sel[sel_key])
    # TODO: decide if we'd actually prefer to call a generalized way of making
    #  ROI FITS in imgops -- but we do want to mess with the individual HDUs
    #  here
    roi_fits = fits.HDUList()
    # iterate over eyes and colors to create HDUs -- note that same colors
    # are not necessarily present in each eye
    for eye, sel_image in sel_images.items():
        color_indices = [
            ix for ix in np.unique(sel_image) if ix != sel["erasecolor"]
        ]
        for color_ix in color_indices:
            roi_fits.append(
                roi_hdu_from_sel(sel_image, color_ix, eye, instrument, sel_fn)
            )
    return roi_fits


def add_merspect_colors_to_edgemaps(edgemap_dict: Mapping):
    for roi_color_eye, _ in edgemap_dict.items():
        color = roi_color_eye.replace(" right", "").replace(" left", "")
        edgemap_dict[roi_color_eye]["color"] = MERSPECT_COLOR_MAPPINGS[color]
    return edgemap_dict



"""we love dividing by zero"""
import warnings

POINTLESS_WARNINGS = (
    r".*divide by zero",
    r".*invalid value enc",
    r".*All-NaN",
    r".*are nearly identical",
    r".*overflow encountered",
)


def divide_by_zero():
    for msg in POINTLESS_WARNINGS:
        warnings.filterwarnings(message=msg, action="ignore")



"""
This module contains utilities for composing lightweight image processing
pipelines intended principally for generating "quicklook"-type browse images
from multispectral data.

Its centerpiece is the class `Look`. `Look` objects iteratively apply a
sequence of functions to one or more input arrays. This module also implements
a simple DSL intended to allow users to modify the quicklook behavior of an
application by editing external configuration files.
"""
from __future__ import annotations

from abc import ABC
from collections.abc import Callable, Mapping, Sequence
from functools import partial
from inspect import signature
from pathlib import Path
from typing import (
    Any,
    Collection,
    Hashable,
    Literal,
    Optional,
    NotRequired,
    TYPE_CHECKING,
    TypedDict,
    Union,
)
import warnings

from dustgoggles.composition import Composition
from dustgoggles.func import argstop, triggerize

import marslab.spectops
from marslab.imgops.imgutils import crop_all, map_filter

if TYPE_CHECKING:
    from matplotlib.figure import Figure
    import numpy as np
    import pandas as pd
    from PIL.Image import Image

STEP_NAMES = (
    "crop",
    "prefilter",
    "mask",
    "look",
    "limiter",
    "postfilter",
    "plotter",
    "bang",
)
"""
Valid steps of instructions in the DSL. Look treats each of these differently
when constructing itself with Look.compile_from_instruction(). Note that the
order of these names is syntactic; do not rearrange the elements of STEP_NAMES.
"""
# noinspection PyTypeHints
LookName = Literal[
    Literal[marslab.spectops.SPECTOP_NAMES],
    Literal[
        "enhanced color", "true color", "dcs", "composite", "nested_composite"
    ],
]
"""
Recognized names of operations for the 'look' step. (If none are appropriate, 
a DSL instruction can instead explicitly specify a Python function.)
"""


class MaskFill(TypedDict):
    """
    Format of the optional 'colorfill' value of a sub-instruction in a 'mask'
    step.
    """

    # what color should we fill the masked regions with?
    # expressed either as a single 0-255 integer (for a grayscale fill))
    # or a tuple of 3 0-255 integers defining the fill's color as RGB.
    color: Union[int, tuple[int, int, int]]
    # alpha / opacity for the color fill layer, expressed as a number between
    # 0 and 1. 0 specifies a fully transparent fill (which is pointless!);
    # 1 specifies a fully opaque fill.
    mask_alpha: float


class LayerDef(TypedDict):
    """
    Format of an element of the 'layers' list that is an optional insert to
    a 'plotter' step. Only relevant if the plotter function itself understands
    it.
    """

    layer_ix: int
    image: "np.ndarray"


# NOTE: This is written in a weird way because 'pass' is a reserved keyword.
MaskInstruction = TypedDict(
    "MaskInstruction",
    {
        # the primary masking function
        "function": Callable,
        "params": NotRequired[Mapping[str, Any]],
        # specifies a color fill for the mask in downstream
        "colorfill": NotRequired[MaskFill],
        # send the mask as an insert to the plotter step?
        "send": NotRequired[bool],
        # pass the mask on to the next function?
        "pass": NotRequired[bool],
    },
)
"""
'mask' steps defined in the DSL cause the generated Look to construct one
or more masks in sequence. This is the format of an sub-instruction in
a 'mask' step.
"""


class StepDef(TypedDict):
    """Format of most steps in a Look DSL instruction."""

    # explicitly-given Python function
    function: Callable
    # kwargs to pass to `function``
    params: NotRequired[Mapping[str, Any]]


class PrefilterDef(StepDef):
    """Format of prefilter step."""

    # if True, assumes the first argument is a tuple of arrays, and applies the
    # prefilter function separately to each.
    map: NotRequired[bool]


class BangDef(TypedDict):
    function: Callable[[], Any]
    params: NotRequired[Mapping[str, Any]]


class LookInstruction(TypedDict):
    """
    Look DSL instructions are Python Mappings that follow this structure.
    """

    # defines a crop on the input image in pixels: (left, right, bottom, top)
    crop: NotRequired[tuple[int, int, int, int]]
    # a first-pass preprocessing step
    prefilter: NotRequired[PrefilterDef]
    # an additional preprocessing step
    limiter: NotRequired[StepDef]
    # the primary look operation
    look: Union[LookName, Callable]
    # kwargs for the primary look operation
    params: NotRequired[Mapping[str, Any]]
    # if this Look is passed to BandSet.make_look_set(), this specifies which
    # of the BandSet's bands it should send to the Look. Does nothing in other
    # cases.
    bands: NotRequired[Sequence[Hashable]]
    # TODO: it's a silly hack to have a required 'instructions' key;
    #  'mask' should just be a sequence of MaskInstructions.
    mask: NotRequired[
        Mapping[Literal["instructions"], Sequence[MaskInstruction]]
    ]
    limiter: NotRequired[StepDef]
    plotter: NotRequired[StepDef]
    # optional niladic function called on successful execution of the rest
    # of the Look. can be used for logging or whatever.
    bang: NotRequired[BangDef]
    # optional identifier for look
    name: NotRequired[str]


# TODO, maybe: we need complicated Protocol stuff to precisely type-hint this
def look_to_function(look: str) -> Callable:
    """
    Implements a mapping between plain-language names of core look operations
    and Python functions.
    """
    from marslab.imgops import render

    if look in marslab.spectops.SPECTOP_NAMES:
        return partial(
            render.spectop_look, spectop=getattr(marslab.spectops, look)
        )
    elif look in ("enhanced color", "true color", "composite"):
        return render.render_rgb_composite
    elif look == "dcs":
        return render.decorrelation_stretch
    elif look == "nested_composite":
        return render.render_nested_rgb_composite
    raise ValueError("unknown look operation " + look)


def interpret_look_step(instruction: LookInstruction):
    """
    Heart of look pipeline: the look function (spectrum op, dcs, etc.), by
    name or as an explicitly-defined Python function. This step is _mandatory_.
    """
    try:
        step = instruction["look"]
    except KeyError:
        raise ValueError("The instruction must include at least a look type.")
    if isinstance(step, str):
        step = look_to_function(step)
    elif callable(step) is False:
        raise TypeError(
            "An instruction's 'look' must be a Python function or a string."
        )
    return step, instruction.get("params", {})


def interpret_crop_step(
    bounds: Optional[tuple[int, int, int, int]]
) -> tuple[Callable, dict]:
    """
    Interpret the crop step of a LookInstruction.
    note at present a cropper is always included whether you ask for it or not,
    but by default it acts as the identity function.
    """
    return crop_all, {"bounds": bounds}


def interpret_prefilter_step(inst: PrefilterDef) -> tuple[Callable, dict]:
    """Interpret the prefilter step of a LookInstruction."""
    step = inst["function"]
    # by default, wrap the prefilter function so that it works as a variadic
    # function, applying itself to each input channel individually; some
    # notional prefilters may of course not want this!
    if inst.get("map") is not False:
        step = map_filter(step)
    # noinspection PyTypeChecker
    return step, inst.get("params", {})


def interpret_mask_step(chunk):
    """Interpret a 'mask' step from a Look instruction."""
    from marslab.imgops.masking import extract_masks

    return extract_masks, chunk


def interpret_instruction_step(
    instruction: LookInstruction, step_name: Literal[STEP_NAMES]
) -> tuple[Optional[Callable], Mapping]:
    """
    unpack an individual element of a LookInstruction into steps, inserts,
    and sends for a Composition.
    """
    # check central mandatory element first
    if step_name == "look":
        return interpret_look_step(instruction)
    # bail out if non-mandatory element is not present
    if step_name not in instruction.keys():
        return None, {}
    if step_name not in STEP_NAMES:
        warnings.warn(
            f"{step_name} is not a known Look step; this may cause unintended "
            f"behavior."
        )
    # noinspection PyTypedDict
    chunk = instruction[step_name]
    # other special cases
    if step_name == "crop":
        return interpret_crop_step(chunk)
    if step_name == "prefilter":
        return interpret_prefilter_step(chunk)
    if step_name == "mask":
        return interpret_mask_step(chunk)
    if step_name == "bang":
        return triggerize(chunk["function"]), chunk.get("params", {}).copy()
    # If no special case is specified, prep for default Composition behavior.
    # specifying a function is mandatory.
    # specifying bound parameters is not.
    return chunk["function"], chunk.get("params", {}).copy()


class Look(Composition, ABC):
    def __init__(
        self,
        *args,
        metadata: "pd.DataFrame" = None,
        bands: tuple[str] = None,
        special_constants: Collection[Any] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.metadata = metadata
        self.bands = bands
        self.special_constants = special_constants
        if (self.metadata is not None) or (self.special_constants is not None):
            self.populate_kwargs()

    def _add_wavelengths(self, wavelengths: Sequence[float]) -> bool:
        """
        Attempt to add an insert to the look step giving it wavelengths. If -
        the look step isn't a spectop, don't actually do this.
        Returns True if the look step is a spectop, and False if not.
        """
        look = self.steps["look"]
        if "__name__" in dir(look):
            name = look.__name__
        else:
            name = look.func.__name__
        if name != "spectop_look":
            return False
        self.add_insert("look", "wavelengths", wavelengths)
        return True

    def _get_plotter_layers(self) -> list[LayerDef]:
        """
        If it exists, return the list of layers prepped as an insert to
        the plotter step. If not, construct an empty one and return it.
        """
        try:
            return self.inserts["plotter"]["layers"]
        except KeyError:
            self.add_insert("plotter", "layers", [])
            return self._get_plotter_layers()

    def add_underlay(self, underlay: "np.ndarray", layer_ix: int = -1):
        """
        Add an underlay to the plotter step. Will only do
        anything if the plotter step understands/accepts layers.
        """
        if "crop" in self.steps:
            underlay = self.steps["crop"](
                underlay, **self.inserts.get("crop", {})
            )
        self._get_plotter_layers().append(
            {"layer_ix": layer_ix, "image": underlay}
        )

    @classmethod
    def compile_from_instruction(
        cls,
        instruction: LookInstruction,
        metadata: "pd.DataFrame" = None,
        # TODO: is this cruft?
        special_constants: Collection[Any] = None,
    ):
        """
        Compile an instruction written in the Look DSL into a rendering
        pipeline. See the docstring of `LookInstruction` for syntax.

        `metadata` is an optional DataFrame formatted like the `metadata`
        property of a `Bandset`. Some `Look` functionality will not work
        without metadata.
        """
        steps = {}
        inserts = {}
        for step_name in STEP_NAMES:
            step, kwargs = interpret_instruction_step(instruction, step_name)
            if step is not None:
                steps[step_name] = step
            if kwargs != {}:
                inserts[step_name] = kwargs
        look = cls(
            steps,
            inserts=inserts,
            metadata=metadata,
            special_constants=special_constants,
            bands=instruction.get("bands"),
        )
        if "mask" in steps.keys():
            # add 'splitter'
            # TODO, maybe: implement this at Composition level
            look.steps["look"] = argstop(look.steps["look"], 2)
            look.add_send("mask", lambda s: s[2], look._get_plotter_layers())
        return look

    # TODO: is this excessively baroque; would an internal dispatch be better?
    def populate_kwargs(self):
        """
        Helper function for Look construction. Populates `Composition`
        structure implied by `Look.steps` and `Look.metadata`.
        """
        for step in self.steps:
            params = signature(self.steps[step]).parameters.values()
            param_names = [param.name for param in params]
            for thing in ("special_constants", "metadata"):
                if (hasattr(self, thing)) and (thing in param_names):
                    self.add_insert(step, thing, getattr(self, thing))
        if self.bands is not None:
            if "WAVELENGTH" in self.metadata.columns:
                wavelengths = []
                for band in self.bands:
                    try:
                        wavelengths.append(
                            self.metadata.loc[
                                self.metadata["BAND"] == band, "WAVELENGTH"
                            ].iloc[0]
                        )
                    except (KeyError, IndexError):
                        continue
                self._add_wavelengths(wavelengths)


# TODO: doesn't go in this module.
def save_plainly(
    look: Union["Figure", "Image"],
    filename: Union[str, Path],
    outpath: Union[str, Path],
    dpi: int = 275,
):
    """
    Save a PIL Image or matplotlib Figure as simply as possible to
    `filename`/`outpath` with specified `dpi`. dpi is ignored if `look` is
    an `Image`.
    """
    from matplotlib.figure import Figure

    if isinstance(look, Figure):
        for ix, axis in enumerate(look.axes):
            if ix > 0:
                axis.remove()
            else:
                axis.axis("off")
        look.savefig(
            Path(outpath, filename), dpi=dpi, bbox_inches="tight", pad_inches=0
        )
    else:
        look.save(Path(outpath, filename))



from functools import reduce

from cytoolz import keyfilter, keymap
from dustgoggles.composition import Composition
from marslab.imgops.imgutils import (
    nanmask,
    ravel_valid,
    cut_annulus,
    cut_rectangle,
)
import numpy as np
from scipy.fft import fft2, fftshift, ifft2, ifftshift
from scipy import stats
from scipy.ndimage import convolve, sobel


def pick_component(complex_array, component="both"):
    if component == "both":
        return complex_array
    if component == "real":
        return complex_array.real
    if component == "imag":
        return complex_array.imag
    if component in ("norm", "magnitude"):
        return np.abs(complex_array)
    if component == "phase":
        return np.arctan2(complex_array.real, complex_array.imag)
    raise ValueError(f"{component} is not a known component.")


def frequency(array, component="both"):
    freq_array = fftshift(fft2(array))
    return pick_component(freq_array, component)


def space(array, component="both"):
    space_array = ifft2(ifftshift(array))
    return pick_component(space_array, component)


def bandfilter(array, bounds, region, component):
    freq_array = frequency(array)
    annulus = cut_annulus(freq_array, bounds, region)
    annulus[annulus.mask] = 0
    return space(annulus.data, component)


def bandcut(array, bounds, component="real"):
    return bandfilter(array, bounds, "outer", component)


def bandpass(array, bounds, component="real"):
    return bandfilter(array, bounds, "inner", component)


def boxfilter(array, bounds, region, component, center=True):
    freq_array = frequency(array)
    rect = cut_rectangle(freq_array, bounds, region, center)
    rect[rect.mask] = 0
    return space(rect.data, component)


def recursive_cut(
    array, cut_method, boundaries, op=np.logical_and, **cut_kwargs
):
    cuts = [
        cut_method(array, bound, **cut_kwargs).mask for bound in boundaries
    ]
    mask = reduce(op, cuts)
    return np.ma.MaskedArray(array, mask=mask)


def fftplot_factory(title=None, component="norm"):
    from marslab.imgops.pltutils import clipshow_factory

    clipshow = clipshow_factory(title)
    transformer = Composition((frequency, clipshow.execute))
    transformer.add_insert(0, "component", component)


def neighbor_kernels():
    kernels = []
    for neighbor_ix in (0, 1, 2, 3, 5, 6, 7, 8):
        kernel = np.zeros((9,))
        kernel[neighbor_ix] = -1
        kernel[4] = 1
        kernels.append(kernel.reshape(3, 3))
    return kernels


def autocompare(image, copy=True):
    abs_image = np.abs(image)
    statistics = {
        "std": np.std(image),
        "mean": np.mean(image),
        "std_abs": np.std(abs_image),
        "mean_abs": np.mean(abs_image),
    }
    edge_results = []
    canvas = nanmask(image, copy)
    for kernel in neighbor_kernels():
        edge_results.append((convolve(canvas, kernel)))
    del canvas
    edge_values = np.abs(ravel_valid(np.concatenate(edge_results)))
    statistics |= {
        "edge_mean_abs": np.mean(edge_values),
        "edge_std_abs": np.std(edge_values),
    }
    sobel_values = np.abs(ravel_valid(sobel(image)))
    statistics |= {
        "sobel_mean_abs": np.mean(sobel_values),
        "sobel_std_abs": np.std(sobel_values),
    }
    return statistics


def center_and_scale(array):
    return (array - array.mean()) / array.std()


def center(array):
    return array - array.mean()


def scale(array):
    return array / array.std()


def unpack_scipy_describe(result):
    return {
        "mean": result.mean,
        "n": result.nobs,
        "min": result.minmax[0],
        "max": result.minmax[1],
        "range": result.minmax[1] - result.minmax[0],
        "skew": result.skewness,
        "kurtosis": result.kurtosis,
        "var": result.variance,
    }


# noinspection PyTypeChecker
def rvdescribe(array, rv=True):
    result = {"shape": array.shape}
    if rv is True:
        array = ravel_valid(array)
    result |= unpack_scipy_describe(stats.describe(array))
    result["std"] = np.sqrt(result["var"])
    if result["min"] >= 0:
        return result
    absresult = unpack_scipy_describe(stats.describe(np.abs(array)))
    absresult = keyfilter(lambda k: k != "n", absresult)
    absresult = keymap(lambda x: f"{x}_abs", absresult)
    absresult["std_abs"] = np.sqrt(absresult["var_abs"])
    return result | absresult


def gradient_stats(
    array, sample_distance=1, get_mag=True, return_components=True
):
    dy, dx = np.gradient(array, sample_distance)
    component_dict = {"dy": dy, "dx": dx}
    if get_mag is True:
        component_dict["norm"] = dy ** 2 + dx ** 2
    stat_dict = {k: rvdescribe(v) for k, v in component_dict.items()}
    if return_components is True:
        return stat_dict, component_dict
    return stat_dict, {}


def fhplot(
    array, bins=None, vrange=None, ax=None, return_counts=False, **mpl_kwargs
):
    try:
        import fast_histogram as fh
        import matplotlib.pyplot as plt
    except ImportError:
        raise ValueError(
            "The fast-histogram and maplotlib libraries must be installed to "
            "use marslab.imgops.imgstats.fhplot."
        )
    if bins is None:
        # TODO, maybe: inefficient; allow passing precalc
        n_unique = len(np.unique(array))
        max_exp = 8
        for i in range(max_exp):
            bins = 2 ** i
            if bins > n_unique:
                break
    if len(array.shape) > 1:
        raise ValueError("this function only plots 1D arrays.")
    if vrange is None:
        vrange = (array.min(), array.max() * 1.001)
    counts = fh.histogram1d(array, bins=bins, range=vrange)
    bin_positions = np.linspace(*vrange, bins + 1)
    if ax is None:
        ax = plt.gca()
    mpl_kwargs = {"histtype": "step"} | mpl_kwargs
    ax.hist(bin_positions[:-1], bin_positions, weights=counts, **mpl_kwargs)
    if return_counts is True:
        return ax, counts, bin_positions
    return ax



from pathlib import Path

from hypothesis import given
import numpy as np
import pdr

from marslab.geom import (
    get_coordinates,
    get_coordinate_system_properties,
    invert_quaternion,
    quaternion_multiplication,
    rotate_unit_vector,
    transform_angle,
)
from marslab.tests.utilz.utilz import quaternions


def test_vicarlike_coordinate_system_transformations():
    data = pdr.read(Path(__file__).parent / "data" / "coords.lbl")
    # perform managed transformation of solar direction vector from ROVER
    # frame to SITE frame
    site_alt, site_az, _ = transform_angle("ROVER", "SITE", "SOLAR", data)
    # get the quaternion that will rotate SITE frame to ROVER frame
    reverse_rot = invert_quaternion(
        get_coordinate_system_properties("ROVER", data)["quaternion"]
    )
    # rotate the computed solar direction vector back to ROVER frame
    rov_el, rov_az, _ = rotate_unit_vector(site_alt, site_az, reverse_rot)
    # compare it to vector explicitly given in the label
    rov_coords = get_coordinates(data)["ROVER"]["SOLAR"]
    assert np.isclose(rov_el, rov_coords["ELEVATION"])
    assert np.isclose(rov_az, rov_coords["AZIMUTH"])


@given(a=quaternions, b=quaternions, c=quaternions)
def test_quaternion_multiplication_associativity(a, b, c) -> None:
    left = quaternion_multiplication(
        q1=a, q2=quaternion_multiplication(q1=b, q2=c)
    )
    right = quaternion_multiplication(
        q1=quaternion_multiplication(q1=a, q2=b), q2=c
    )
    assert np.allclose(left, right, atol=1e-5)


@given(quat=quaternions)
def test_quaternion_identity_element(quat) -> None:
    identity = [1, 0, 0, 0]
    assert (quat == quaternion_multiplication(quat, identity)).all()
    assert (quat == quaternion_multiplication(identity, quat)).all()



import warnings
from functools import reduce
from operator import and_

import numpy as np

from marslab.imgops.regions import count_rois_on_image
from marslab.tests.utilz.div0 import divide_by_zero

rng = np.random.default_rng()


# TODO: parametrize, etc.


def test_with_trivial_case():
    length = rng.integers(25, 500)
    tile_size = (length // 2, length // 2)
    blackwhite_top = np.hstack([np.zeros(tile_size), np.ones(tile_size)])
    blackwhite = np.vstack([blackwhite_top, np.fliplr(blackwhite_top)])
    rois = {ix: np.full(blackwhite.shape, False) for ix in range(4)}
    rois[0][0 : length // 2, 0 : length // 2] = True
    rois[1][0 : length // 2, length // 2 :] = True
    rois[2][length // 2 :, 0 : length // 2] = True
    rois[3][length // 2 :, length // 2 :] = True
    blackwhite[length // 2 + 1, 0] = -9999
    with warnings.catch_warnings():
        divide_by_zero()
        counts = count_rois_on_image(
            rois.values(), rois.keys(), blackwhite, special_constants=[-9999]
        )
    assert reduce(
        and_,
        (
            counts[0]["total"] == 0,
            counts[1]["total"] == blackwhite.size / 4,
            counts[2]["total"] == blackwhite.size / 4 - 1,
            counts[3]["total"] == 0,
        ),
    )



"""
image processing utility functions
"""
import gc
import sys
from functools import partial, reduce
from itertools import chain
from numbers import Number, Real
from operator import mul
from typing import (
    Callable,
    Sequence,
    MutableMapping,
    Collection,
    Mapping,
    Union,
    MutableSequence,
    Any,
    Hashable,
    Optional,
    Literal,
)

from dustgoggles.structures import dig_for_values
import numpy as np
from numpy.typing import ArrayLike


# TODO: maybe does not belong in this module
# TODO, maybe: add even more stuff -- try to run .close(), etc.
def absolutely_destroy(
    thing: Union[MutableMapping, MutableSequence], delay_collect: bool = False
) -> None:
    """
    Object-clearing utility intended primarily but not exclusively for mutable
    collections of matplotlib objects.

    Clear the contents of `thing`, which may be a list-like or dict-like
    object; then, close all matplotlib figures (if matplotlib.pyplot is
    loaded); then (if `delay_collect` is False), run gc.collect().

    This throw-things-at-the-wall approach is intended to trigger various
    destructor behaviors for collections and elements of collections that may
    not respond as expected to methods such as .clear() and .close(), and may
    enjoy creating reference cycles.
    """
    if isinstance(thing, MutableMapping):
        keys = list(thing.keys())
        for key in keys:
            del thing[key]
    elif isinstance(thing, MutableSequence):
        [thing.pop() for _ in range(len(thing))]
    del thing
    if "matplotlib.pyplot" in sys.modules.keys():
        getattr(sys.modules["matplotlib.pyplot"], "close")("all")
    if delay_collect is False:
        gc.collect()


def eightbit(
    array: np.ndarray, stretch: Union[float, tuple[float, float], None] = None
) -> np.ndarray:
    """
    Return an 'eight-bit' (single-byte unsigned integer, uint8, u1, etc.)
    version of an array scaled to (0, 255); optionally, stretch (range-clip)
    it inline. Intended primarily for preparing arrays to be interpreted as
    images (or image channels) in a conventional colorspace of some kind.

    If `array` is already uint8, does not modify it unless stretch is not None,
    in which case it will map its min/max values to 0/255 (optionally
    stretched) as with any other input array.

    WARNING: If `array` is a MaskedArray, this function automatically fills
      corresponding masked values of the output array with 0s to prevent
      undesirable typecasting behaviors.
    """
    if array.dtype.char == "B" and stretch is None:
        return array
    if array.dtype.char == "B":
        return normalize_range(array, (0, 255), stretch)
    output = normalize_range(array, (0, 255), stretch)
    if isinstance(output, np.ma.MaskedArray):
        output[output.mask] = np.uint8(0)
    if output.dtype.char not in np.typecodes["AllInteger"]:
        output = np.round(output)
    return output.astype(np.uint8)


def cropmask(
    arr: np.ndarray,
    bounds: Optional[tuple[int, int, int, int]] = None,
    copy: bool = True,
    **_,
) -> np.ndarray:
    """
    Return a MaskedArray based on `arr` with all elements that fall within a
    rectangular 'border' masked (along with any already-masked elements, if
    `arr` is a MaskedArray).
    `bounds` is a tuple of  (left, right, top, bottom) pixels to crop, or
    None; if `bounds` is None; simply return `arr`.

    If `copy` is False and `arr` is a MaskedArray, modify `arr.mask` inplace.
    The `copy` argument has no effect if `arr` is not a MaskedArray.
    """
    if bounds is None:
        return arr
    assert len(bounds) == 4  # test for bad inputs
    pixels = [side if side != 0 else None for side in bounds]
    canvas = np.ones(arr.shape)
    for value in (1, 3):
        if isinstance(pixels[value], int):
            if pixels[value] > 0:
                pixels[value] = pixels[value] * -1
    canvas[pixels[2] : pixels[3], pixels[0] : pixels[1]] = 0
    if isinstance(arr, np.ma.MaskedArray):
        if copy is True:
            arr.mask = np.logical_or(canvas, arr.mask)
            return arr
        return np.ma.masked_array(arr.data, np.logical_or(canvas, arr.mask))
    return np.ma.masked_array(arr, canvas)


def crop(
    arr: np.ndarray, bounds: Optional[tuple[int, int, int, int]] = None, **_
) -> np.ndarray:
    """
    return a copy of `arr` cropped to specified rectangular bounds.
    `bounds` is a tuple of (left, right, top, bottom) pixels to crop, or None;
    if `bounds` is None, simply return `arr`.
    """
    if bounds is None:
        return arr
    assert len(bounds) == 4  # test for bad inputs
    pixels = [side if side != 0 else None for side in bounds]
    for value in (1, 3):
        if isinstance(pixels[value], int):
            if pixels[value] > 0:
                pixels[value] = pixels[value] * -1
    return arr[pixels[2] : pixels[3], pixels[0] : pixels[1]]


def crop_all(
    arrays: Collection[np.ndarray], bounds=None, **_
) -> Union[Mapping[str, list[np.ndarray]], list[np.ndarray], np.ndarray]:
    """applies crop() to every array in `arrays`."""
    # if you _didn't_ pass it a collection, just overload / dispatch to crop()
    if isinstance(arrays, np.ndarray):
        return crop(arrays, bounds)
    # otherwise map crop()
    elif isinstance(arrays, Mapping):
        return {
            name: [crop(array, bounds) for array in name_arrays]
            for name, name_arrays in arrays.items()
        }
    return [crop(array, bounds) for array in arrays]


def clip_unmasked(arr: np.ma.MaskedArray) -> np.ndarray:
    """
    np.clip() does not always respect masks properly. Return an unmasked
    version of `arr` clipped to the min/max values of its unmasked elements.
    """
    return np.clip(arr.data, arr.min(), arr.max())


# TODO: this may no longer be necessary for some functions due to upstream
#  API changes (notably, the referred-to gaussian blur function).
def split_filter(
    filter_function: Callable[[np.ndarray, ...], np.ndarray], axis: int = -1
) -> Callable:
    """
    produce a 'split' version of a function -- notionally an image filter --
    that applies itself to slices across a particular axis. e.g., if passed a
    gaussian blur function, returns a function that applies a gaussian blur to
    R / G / B channels separately and then recomposes them.
    """

    def multi(
        array: np.ndarray, *, set_axis: int = axis, **kwargs: Any
    ) -> np.ndarray:
        filt = partial(filter_function, **kwargs)
        filtered = tuple(
            map(filt, np.split(array, array.shape[set_axis], set_axis))
        )
        if isinstance(filtered[0], np.ma.MaskedArray):
            return np.ma.concatenate(filtered, axis=set_axis)
        return np.concatenate(filtered, axis=set_axis)

    return multi


def map_filter(filter_function: Callable) -> Callable:
    """
    returns a version of a function -- notionally an image filter -- that
    automatically maps itself across all elements of a passed collection.
    """

    def mapped_filter(arrays, *args, **kwargs):
        return [filter_function(array, *args, **kwargs) for array in arrays]

    return mapped_filter


def find_masked_bounds(
    image: np.ma.MaskedArray, cheat_low: float, cheat_high: float
) -> Union[tuple[Real, Real], tuple[None, None]]:
    """
    relatively memory-efficient way to perform bound calculations for
    normalize_range on a masked array.
    """
    if image.mask is np.False_:
        valid = image.data
    else:
        valid = image[~image.mask].data
    # NOTE: unlike its relative in `pdr.browsify`, this function does not
    #  assume that the masked array it receives has already been masked
    #  where invalid.
    valid = valid[np.isfinite(valid)]
    if valid.size == 0:
        return None, None
    if (cheat_low != 0) and (cheat_high != 0):
        minimum, maximum = np.percentile(
            valid, [cheat_low, 100 - cheat_high], overwrite_input=True
        ).astype(image.dtype)
    elif cheat_low != 0:
        maximum = valid.max()
        minimum = np.percentile(valid, cheat_low, overwrite_input=True).astype(
            image.dtype
        )
    elif cheat_high != 0:
        minimum = valid.min()
        maximum = np.percentile(
            valid, 100 - cheat_high, overwrite_input=True
        ).astype(image.dtype)
    else:
        minimum = valid.min()
        maximum = valid.max()
    return minimum, maximum


# noinspection PyArgumentList
def find_unmasked_bounds(
    image: np.ndarray, cheat_low: float, cheat_high: float
) -> tuple[Real, Real]:
    """straightforward way to find unmasked array bounds for normalize_range"""
    if cheat_low != 0:
        minimum = np.percentile(image, cheat_low).astype(image.dtype)
    else:
        minimum = image.min()
    if cheat_high != 0:
        maximum = np.percentile(image, 100 - cheat_high).astype(image.dtype)
    else:
        maximum = image.max()
    return minimum, maximum


def normalize_range(
    arr: np.ndarray,
    bounds: tuple[Real, Real] = (0.0, 1.0),
    stretch: Union[float, tuple[float, float], None] = None,
    inplace: bool = False,
    allow_half_float: bool = False,
) -> Union[np.ndarray, np.ma.MaskedArray]:
    """
    simple linear min-max scaler that optionally percentile-clips the input at
    stretch = (low_percentile, 100 - high_percentile). if inplace is True,
    may transform the original array, with attendant memory savings and
    destructive effects.

    Setting bounds[0] >= bounds[1] may cause undesired effects.

    Attempts to maintain dtype of the input array, but will cast if required
    to match bounds. Note that the inplace argument is effectively ignored in
    that case.

    Always ignores masked data in an array when computing scale.
    Arrays with nan values will always return an array of np.nan. Arrays with
    -inf and/or inf values will clip those values to bounds[0] and bounds[1]
    respectively.
    """
    if not isinstance(stretch, Sequence):
        stretch = (stretch, stretch)
    stretch = np.array([0 if i is None else i for i in stretch])
    stretch = np.clip(stretch, 0, 100)
    do_clip = max(stretch) > 0
    if isinstance(arr, np.ma.MaskedArray):
        inrange = find_masked_bounds(arr, *stretch)
    else:
        inrange = find_unmasked_bounds(arr, *stretch)
    inrange = tuple(
        map(lambda i: arr.dtype.type(i) if i is not None else i, inrange)
    )
    if do_clip and set(inrange) != {None}:
        if inplace is True:
            arr = np.clip(arr, *inrange, out=arr)
        else:
            arr = np.clip(arr, *inrange)
    mintype = reduce(
        np.promote_types,
        map(np.min_scalar_type, filter(None, (*inrange, *bounds))),
    )
    if mintype == np.float16 and allow_half_float is False:
        mintype = np.float32
    if not np.can_cast(mintype, arr.dtype):
        arr = arr.astype(mintype)
    # Note that this case will only occur for all-masked input array
    if set(inrange) == {None}:
        return np.clip(arr, *bounds)
    inrange, bounds = map(
        lambda s: np.array(s, arr.dtype.type), (inrange, bounds)
    )
    scale_up, scale_down = map(np.ptp, (bounds, inrange))
    # TODO, maybe: messy
    if arr.dtype.char in np.typecodes["Float"]:
        # nonsensical-looking but effective floating point OOB cheat
        smax = (inrange[1] - inrange[0]) * scale_up / scale_down + bounds[0]
        if smax > bounds[1]:
            scale_up *= 1 + (bounds[1] - smax)
    if inplace is True:
        arr -= inrange[0]
        arr *= scale_up
        if arr.dtype.char in np.typecodes["AllInteger"]:
            # this loss of precision is probably better than
            # automatically typecasting it.
            # TODO, maybe: cute inplace rounding
            arr //= scale_down
        else:
            arr /= scale_down
        arr += bounds[0]
        return arr
    elif arr.dtype.char in np.typecodes["AllInteger"]:
        return ((arr - inrange[0]) * scale_up) // scale_down + bounds[0]
    return (arr - inrange[0]) * scale_up / scale_down + bounds[0]


def enhance_color(image: np.ndarray, bounds, stretch):
    """
    wrapper for normalize_range -- normalize each channel individually,
    conventional "enhanced color" operation
    """
    return split_filter(normalize_range)(image, bounds=bounds, stretch=stretch)


# TODO: probably pointless cruft. verify.
def clip_finite(image, a_min: int, a_max: int):
    finite = np.ma.masked_invalid(image)
    result = np.ma.clip(finite, a_min, a_max)
    if isinstance(image, np.ma.MaskedArray):
        return result
    return result.data


def std_clip(arr: np.ndarray, sigma: Number = 1) -> np.ndarray:
    """
    simple clipping function that clips at multiples of an array's standard
    deviation offset from its mean
    """
    finite = np.ma.masked_invalid(arr)
    mean, shift = np.ma.mean(finite), np.ma.std(finite) * sigma
    clips = np.array([mean - shift, mean + shift], dtype=arr.dtype)
    result = np.ma.clip(finite, *clips)
    if isinstance(arr, np.ma.MaskedArray):
        return np.ma.masked_array(result.data, mask=(arr.mask + result.mask))
    # NOTE: np typestubs suggest MaskedArray.data is a memoryview, but it's not
    # noinspection PyTypeChecker
    return result.data


def centile_clip(
    arr: np.ndarray, centiles: tuple[Number, Number] = (1, 99)
) -> np.ndarray:
    """
    simple clipping function that clips values above and below a given
    percentile range
    """
    finite = np.ma.masked_invalid(arr)
    bounds = np.percentile(finite[~finite.mask].data, centiles)
    result = np.ma.clip(finite, *(bounds.astype(arr.dtype)))
    if isinstance(arr, np.ma.MaskedArray):
        return np.ma.masked_array(result.data, mask=(arr.mask + result.mask))
    # NOTE: np typestubs suggest MaskedArray.data is a memoryview, but it's not
    # noinspection PyTypeChecker
    return result.data


def minmax_clip(arr: np.ndarray, stretch: tuple[float, float] = (0, 0)):
    """
    simple minmax clip to cheat min and max up and down at multiples of an
    array's dynamic range. always returns a copy.
    """
    if max(stretch) == 0:
        return arr.copy()
    if max(stretch) > 1:
        raise ValueError("Stretch bounds must be between 0 and 1 inclusive.")
    finite = np.ma.masked_invalid(arr)
    fmin, fmax = finite.min(), finite.max()
    ptp = fmax - fmin
    result = np.ma.clip(
        finite,
        finite.dtype.type(fmin + stretch[0] * ptp),
        finite.dtype.type(fmax - stretch[1] * ptp),
    )
    if isinstance(arr, np.ma.MaskedArray):
        return np.ma.masked_array(result.data, mask=(arr.mask + result.mask))
    return result.data


def sinh_scale(
    arr: np.ndarray,
    bounds: tuple[Number, Number] = (0.0, 1.0),
    stretch: float = 1,
) -> np.ndarray:
    """
    Applies a simple hyperbolic sine stretch to the input array. Maintains
    the dtype of the input array if it is floating-point; otherwise, converts
    it to float64.
    """
    finite = np.ma.masked_invalid(arr)
    fmin, fmax = finite.min(), finite.max()
    scaling = 1 / (fmax - fmin)
    if arr.dtype.char in np.typecodes["Float"]:
        scaling = arr.dtype.type(scaling)
    normed = normalize_range(
        np.sinh((finite - fmin) * scaling * stretch - 0.5), bounds
    )
    if isinstance(arr, np.ma.masked_array):
        return np.ma.masked_array(normed, mask=normed.mask + arr.mask)
    # noinspection PyTypeChecker
    return normed.data


# TODO: would it be better for these functions to use flat mask arrays
#  rather than coordinate grids? (performance is the same, question is
#  interface convenience)
def bilinear_interpolate(
    input_array: np.ndarray,
    output_shape: tuple[int, int],
    y_coords: np.ndarray,
    x_coords: np.ndarray,
) -> np.ndarray:
    """
    upsample a 2D array, gridding its pixels at y-coordinates given in y_coords
    and x-coordinates given in x_coords and linearly interpolating from there,
    first in the x direction and then in the y direction. y_coords and x_coords
    are 2D arrays of the same size like those created by np.meshgrid,
    defining locations in N**2
    This is the general (gridded) case and should work correctly for any
    pixel class defined as a partition of a grid. Use the faster
    bilinear_interpolate_subgrid() for pixel classes that can be defined as
    locations relative to a regular tiling of a grid.
    """
    horizontal = np.zeros(output_shape, dtype=input_array.dtype)
    vertical = np.zeros(output_shape, dtype=input_array.dtype)
    rows = np.unique(y_coords)
    for row_ix in rows:
        coordinate_reference_indices = np.where(y_coords == row_ix)
        column_indices = x_coords[coordinate_reference_indices]
        horizontal[row_ix] = np.interp(
            np.arange(output_shape[1]),
            column_indices,
            input_array[coordinate_reference_indices],
        )
    # column indices are immaterial now, because we've interpolated the rows.
    for column_ix in np.arange(output_shape[1]):
        vertical[:, column_ix] = np.interp(
            np.arange(output_shape[0]), rows, horizontal[rows, column_ix]
        )
    return vertical


def bilinear_interpolate_subgrid(
    rows: np.ndarray,
    columns: np.ndarray,
    input_array: np.ndarray,
    output_shape: tuple[int, int],
) -> np.ndarray:
    """
    interpolate 2D values to a 2D array, gridding those values
    according to a regular pattern. input array must be of an
    integer or float dtype.

    this is a special case for pixel classes that are defined as unique
    positions within m x n subgrids that tile the coordinate space.
    in particular, it will work for any conventional Bayer pattern,
    when each pattern cell is treated as a separate pixel 'class' (even if
    pixels are of the 'same' color), as well as a variety of more complex
    patterns. use the slower bilinear_interpolate() for general gridded-data
    cases.
    """
    horizontal = np.empty(output_shape, dtype=input_array.dtype)
    vertical = np.empty(output_shape, dtype=input_array.dtype)
    # because of the 'subgrid' assumption, in any row that contains pixels,
    # their column indices are the same (and vice versa)
    for row_ix, row in enumerate(rows):
        horizontal[row] = np.interp(
            np.arange(output_shape[1]),
            columns,
            input_array[row_ix, :],
        )
    for output_column in np.arange(output_shape[1]):
        vertical[:, output_column] = np.interp(
            np.arange(output_shape[0]), rows, horizontal[rows, output_column]
        )
    return vertical


def apply_image_filter(image, image_filter=None):
    """
    unpacking / pseudo-dispatch function

    `image_filter` is a dict() that contains 'params' and 'function'
    keys. The optional 'params' value is a parameters dict() and the 'function'
    value is the corresponding function to run on `image`.
    """
    if image_filter is None:
        return image
    return image_filter["function"](image, **image_filter.get("params", {}))


# TODO, maybe: doesn't really go in this module
def mapfilter(
    predicate: Callable[[Any], bool],
    key: Optional[Hashable],
    map_sequence: Sequence[Mapping],
) -> list[Mapping]:
    """
    Return a list containing all elements of map_sequence such that, if `key`
    is None, `predicate(element)` is True, or if `key` is not None,
    `predicate(element.get(key))` is True.
    Intended primarily as a helper function for picking looks.
    """
    new_sequence = []
    for mapping in map_sequence:
        obj = mapping if key is None else mapping.get(key)
        if predicate(obj):
            new_sequence.append(mapping)
    return new_sequence


def make_mask_passer(
    func: Callable[[np.ndarray, ...], np.ndarray], mask_nans=True
) -> Callable[[np.ndarray, ...], np.ndarray]:
    """
    Function decorator for functions that take at least a single positional
    ndarray argument. Returns a function that attempts to restore the masks of
    any MaskedArray passed as the first argument after passing it (and all
    other arguments) to the decorated function. Will of course fail if `func`
    mutates the mask inplace or similar. Optionally (and by default), also
    mask NaNs inline.
    """

    def mask_passer(array, *args, **kwargs):
        transformed = func(array, *args, **kwargs)
        if isinstance(array, np.ma.MaskedArray):
            mask = array.mask
            if mask_nans:
                mask[np.isnan(transformed)] = True
            return np.ma.MaskedArray(transformed, mask)
        return transformed

    return mask_passer


# TODO, maybe: doesn't really go in this module.
# TODO, maybe: this probably shouldn't actually return an iterator.
def get_all_bands(instruction: Mapping) -> chain:
    """
    helper function for look set analysis: get all bands mentioned in an
    instructions, including an instructions with nested bands
    """
    return chain.from_iterable(dig_for_values(instruction, "bands"))


# TODO, maybe: doesn't really go in this module.
def get_all_bands_from_all(instructions: Collection[Mapping]) -> set:
    """
    helper function for look set analysis: get all bands mentioned in all
    instructions, including from instructions with nested bands
    """
    return set(chain.from_iterable(map(get_all_bands, instructions)))


def closest_ratio(integer: int, target: float) -> tuple[int, int]:
    """
    Return (x, y) satisfying:

    (a, b ∈ ℕ) ∧ (a * b ==`integer`)
     → abs((x / y) - `target`) <= abs((a / b) - `target`)
    """
    from sympy import factorint

    factors = tuple(
        chain.from_iterable(
            [[f] * count for f, count in factorint(integer).items()]
        )
    )
    factors = tuple(reversed(factors))
    factorizations = []
    possibilities = []
    for unique in set(factors):
        left = [unique]
        right = [f for f in factors if f not in left]
        quantity = factors.count(unique)
        right += [unique] * (quantity - 1)
        ratio = target
        new_left, new_right = left, right
        while ratio <= target:
            for check_ix in range(len(right)):
                new_left = left + [right[check_ix]]
                for f in new_left:
                    if f in new_right:
                        intermediate = [r for r in right if r != f]
                        quantity = right.count(f)
                        intermediate += [f] * (quantity - 1)
                        new_right = intermediate
                ratio = reduce(mul, new_left) / reduce(mul, new_right)
            if ratio <= target:
                left, right = new_left, new_right
        possibilities.append(reduce(mul, left) / reduce(mul, right))
        factorizations.append((left, right))
        possibilities.append(ratio)
        factorizations.append((new_left, new_right))
    differences = [abs(p - target) for p in possibilities]
    closest = min(differences)
    for p, diff, fact in zip(possibilities, differences, factorizations):
        if diff == closest:
            return p, fact


def closest_aspect(size: int, aspect_ratio: float) -> tuple[int, int]:
    """
    Given an array size and a desired aspect ratio, find the integer values
    for width and height that most closely approximate the ratio.
    """
    _, factors = closest_ratio(size, aspect_ratio)
    left, right = factors
    return reduce(mul, right), reduce(mul, left)


def strict_reshape(array: np.ndarray, aspect_ratio: float) -> np.ndarray:
    """
    Return a copy of `array` reshaped such that its aspect ratio is as close
    as possible to `aspect_ratio` (without discarding or adding any elements).
    """
    return array.reshape(closest_aspect(array.size, aspect_ratio))


def ravel_valid(arr: np.ndarray) -> np.ndarray:
    """
    Return a raveled (i.e. flattened, 1-D) copy of `arr` with all 'invalid'
    elements discarded, where 'invalid' means nonfinite and/or (if `arr`
    is a `MaskedArray`) masked.

    Intended primarily as a convenience for aggregate statistical operations,
    particularly operations that use library functions that don't like inf/nan
    and/or don't respect masks.
    """
    raveled = arr.ravel()
    if arr.dtype.char in np.typecodes["AllInteger"] and not isinstance(
        arr, np.ma.MaskedArray
    ):
        # this case can't actually have 'invalid' values and it is a waste of
        # time to check for them
        return raveled
    if isinstance(raveled, np.ma.MaskedArray):
        return raveled[np.isfinite(raveled) + raveled.mask].data
    return raveled[np.isfinite(raveled)]


rv = ravel_valid
"""alias for `ravel_valid`."""


def setmask(arr: np.ndarray, value: Any, copy: bool = True) -> np.ndarray:
    """
    Convenience wrapper/handler for MaskedArray 'filling' behaviors that
    politely ignores non-masked arrays.

    Returns a (non-masked) array like `arr`, but with all masked elements
    replaced with `value`. If `copy`  is True, calls `arr.filled()` (which
    creates a copy of `arr`). Otherwise, modifies `arr.data` inplace and
    returns it.

    If `arr` is _not_ a MaskedArray, simply returns `arr` if `copy` is False,
    and `arr.copy()` if `copy` is True.

    Note: Does not attempt to typecast `arr` to match `value` or vice versa.
     Will throw an exception on numpy>=2.0.0 if `value` is incompatible with
     `arr.dtype`.
    """
    if not isinstance(arr, np.ma.MaskedArray):
        return arr if copy is False else arr.copy()
    if copy is True:
        return arr.filled(value)
    arr.data[arr.mask] = value
    return arr.data


def zero_mask(arr: np.ndarray, copy: bool = True) -> np.ndarray:
    """Alias for `setmask(arr, 0, copy)`"""
    return setmask(arr, 0, copy)


def nanmask(arr: np.ndarray, copy: bool = True) -> np.ndarray:
    """Alias for `setmask(arr, np.nan, copy)`"""
    return setmask(arr, np.nan, copy)


def make_mask_canvas(
    image: np.ma.MaskedArray, unmasked_alpha: float = 1, mask_alpha: float = 0
) -> np.ndarray:
    """
    Creates a half-float array shaped like the first two dimensions of `image`,
    which must be a 2D or 3D ndarray.

    The 'canvas' array's elements are `masked_alpha` where `image.mask` is True
    (True in any 'channel'/'band', if `image` is 3D), and 'unmasked_alpha'
    elsewhere.

    This can be used to create an alpha channel or as a sort of stencil.
    """
    if image.ndim == 3:
        mask = image.mask.sum(axis=-1).astype(bool)
    elif image.ndim == 2:
        mask = image.mask
    else:
        raise ValueError("`image` must be 2D or 3D")
    return np.where(mask, np.float16(mask_alpha), np.float16(unmasked_alpha))


def colorfill_maskedarray(
    masked_array: np.ma.MaskedArray,
    color: Union[float, tuple[float, float, float]] = 0.45,
    mask_alpha=None,
    unmasked_alpha=None,
) -> np.ndarray:
    """
    masked_array: 2-D masked array or a 3-D masked array with last axis of
    length 3. should be float normalized to 0-1.
    color: optionally-specified RGB color or single integer for gray
    (default medium gray).
    if both alpha parameters are None, return a 3-D array with masked values
    filled with color.
    otherwise, return a 4-D array (last axis intended as an alpha plane),
    with alpha value equal to unmasked_alpha outside the mask and masked_alpha
    inside the mask (either defaults to 0 if not set)
    """
    if (mask_alpha is None) and (unmasked_alpha is not None):
        mask_alpha = 0
    if (unmasked_alpha is None) and (mask_alpha is not None):
        unmasked_alpha = 0
    if len(masked_array.shape) == 2:
        if isinstance(color, (int, float)):
            color = (color, color, color)
        stack = [masked_array.filled(color[ix]) for ix in range(3)]
        if mask_alpha is not None:
            stack.append(
                make_mask_canvas(masked_array, unmasked_alpha, mask_alpha)
            )
        return np.dstack(stack)
    if masked_array.shape[-1] != 3:
        raise ValueError("3-D arrays must have last axis of length = 3")
    if isinstance(color, (int, float)):
        stack = masked_array.filled(color)
    else:
        stack = np.dstack(
            [masked_array[:, :, ix].filled(color[ix]) for ix in range(3)]
        )
    if mask_alpha is None:
        return stack
    return np.dstack(
        [stack, make_mask_canvas(stack, unmasked_alpha, mask_alpha)]
    )


def maskwhere(
    images: Sequence[np.ndarray], constants: Collection[ArrayLike]
) -> list[np.ndarray]:
    """
    Return a list of boolean ndarrays corresponding to `images` such that
    the i-th element of that list is True where `images[i]` is in `constants`
    and False elsewhere.

    Intended primarily as a convenience function for producing special constant
    masks for a group of related images.
    """
    # noinspection PyTypeChecker
    return reduce(np.logical_or, [np.isin(i, constants) for i in images])


def _pick_mask_constructors(
    region: Literal["inner", "outer"]
) -> tuple[
    Callable[[np.ndarray, Any, Any], np.ma.MaskedArray],
    Callable[[np.ndarray, np.ndarray], np.ndarray],
]:
    """
    Helper function for cut_annulus() and cut_rectangle(). Selects a relation
    and a logical operator that facilitate masking everything outside or inside
    a set of bounds.
    """
    if region == "inner":
        return np.ma.masked_outside, np.logical_or
    elif region == "outer":
        return np.ma.masked_inside, np.logical_and
    raise ValueError(f"region={region}; region must be 'inner' or 'outer'")


def _make_cut_mask(
    distance: np.ndarray,
    bounds: tuple[Real, Real],
    region: Literal["inner", "outer"],
) -> np.ndarray:
    """
    Generate a mask based on whether elements of `distance` fall within
    specified bounds.
    """
    mask_method, _ = _pick_mask_constructors(region)
    pass_min = bounds[0] if bounds[0] is not None else distance.min()
    pass_max = bounds[1] if bounds[1] is not None else distance.max()
    return mask_method(distance, pass_min, pass_max).mask


def centered_indices(arr: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    """
    Return a pair of arrays giving offsets from center, in index units, for
    the 2D array `arr` (or any other array of its shape).

    NOTE: These cannot actually be used as index arrays! They represent
        'distance' along each axis from center.
    """
    if arr.ndim != 2:
        raise ValueError("This function only takes 2D arrays.")
    y, x = np.indices(arr.shape)
    y0, x0 = (arr.shape[0] - 1) / 2, (arr.shape[1] - 1) / 2
    return y - y0, x - x0


def radial_index(arr: np.ndarray) -> np.ndarray:
    """
    Return an array giving Euclidean distance from center, in index units, for
    the 2D array `arr` (or any other array of its shape).
    """
    y_ix, x_ix = centered_indices(arr)
    return np.sqrt(y_ix ** 2 + x_ix ** 2)


def join_cut_mask(
    arr: np.ndarray, cut_mask: np.ndarray, copy: bool = True
) -> np.ma.MaskedArray:
    """
    Convenience function that masks `arr` with `cut_mask` if `arr` is not a
    `MaskedArray`, or adds `cut_mask` to `arr.mask` if `arr` is a MaskedArray.

    If `copy` is False and `arr` is a MaskedArray, this modifies `arr.mask`
    inplace. `copy` has no effect if `arr` is not a MaskedArray.

    `cut_mask` must be the same shape as `arr`; for best results, it should be
    boolean.
    """
    if not isinstance(arr, np.ma.MaskedArray):
        return np.ma.MaskedArray(arr, mask=cut_mask)
    if copy is True:
        arr = arr.copy()
    arr.mask = np.logical_or(cut_mask, arr.mask)
    return arr


def cut_annulus(
    arr: np.ndarray,
    bounds: tuple[Real, Real],
    region: Literal["inner", "outer"] = "inner",
    copy: bool = True,
) -> np.ma.MaskedArray:
    """
    Return a version of the 2D ndarray `arr` with all values
    inside or outside of an annulus masked. `bounds[0]` defines the inner
    bound of the annulus; `bounds[1]` the outer. These definitions are given
    in units of array-index distance (which is to say 'pixels' or
    'picture-plane distance' for most images). The Euclidean norm is assumed.

    For example:
    `cut_annulus(arr, (100, 200), 'inner')` will return a version of `arr` that
    is masked everywhere except within a ring bounded by 100 array-index units
    from the center of `arr` and 200 array-index units from the center of
    `arr`. `cut_annulus(arr, (100, 200), 'outer')` will do the same, except
    that the array it returns will be masked _only_ within that ring.
    """
    dist = radial_index(arr)
    return join_cut_mask(arr, _make_cut_mask(dist, bounds, region), copy)


def cut_rectangle(
    arr: np.ndarray,
    bounds: tuple[tuple[Real, Real], tuple[Real, Real]],
    region: Literal["inner", "outer"] = "inner",
    center: bool = True,
    copy: bool = True,
) -> np.ma.MaskedArray:
    """
    Return a version of the 2D ndarray `arr` with all values
    inside or outside of a rectangle masked. `bounds` must be a tuple of
    two tuples, each of which contains two numbers defining extrema of the
    rectangle along the x and y axes respectively. These definitions are given
    in units of array-index distance (which is to say 'pixels' for most
    images). If `center` is True (which it is by default), `bounds` defines
    edges with reference to offsets from the center of `arr`. If `center` is
    False, `bounds` defines edges with reference to the upperleft element of
    `arr` (in other words, standard numpy indexing).

    For example:
    `cut_rectangle(arr, ((-100, 100), (-200, 100)), 'inner')` will return a
    version of `arr` that is masked everywhere except within a rectangle
    that extends 100 pixels left, 100 pixels right, 200 pixels down, and 100
    pixels up from the its center element.
    """
    dists = centered_indices(arr) if center is True else np.indices(arr.shape)
    masks = [
        _make_cut_mask(dist, bound, region)
        for bound, dist in zip(bounds, dists)
    ]
    cut_mask = _pick_mask_constructors(region)[0](*masks)
    return join_cut_mask(arr, cut_mask, copy)


def zerocut(
    arr: np.ndarray,
    bounds: tuple[tuple[Real, Real], tuple[Real, Real]],
    region: Literal["inner", "outer"] = "inner",
    center: bool = True,
    copy: bool = True,
) -> np.ndarray:
    """
    Convenience wrapper for `cut_rectangle()` that automatically sets all
    masked elements to 0 and drops the mask. Intended primarily for quick
    renders and spectral analyses.
    """
    return zero_mask(cut_rectangle(arr, bounds, region, center, copy))



"""skeleton BandSet for MSL Mastcam multispectral observations"""

from functools import cache, partial
from operator import attrgetter
from pathlib import Path
from types import MappingProxyType

import pandas as pd
from dustgoggles.scrape import bulk_scrape_metadata, scrape_patterns

from marslab.compat.xcam import DERIVED_CAM_DICT, BAND_TO_BAYER
from marslab.bandset.bandset import BandSet
from marslab.imgops.debayer import RGGB_PATTERN
from marslab.imgops.loaders import pdr_load


def parse_mcam_fn(mcam_fn):
    filename = Path(mcam_fn).name
    return {
        "EYE": filename[5],
        "SEQ_ID": f"mcam{int(filename[6:12])}",
        "CAL": filename[26:30],
        "FILETYPE": filename[22:25],
    }


MCAM_METADATA_REGEX = MappingProxyType(
    {
        "CSEQ": r"COMMAND_SEQUENCE_NUMBER.*?(\d+)",
        "SOL": r"(?<=PLANET_DAY_NUMBER).*?(\d+)",
    }
)

scrape_mcam_metadata = cache(
    partial(scrape_patterns, metadata_regex=MCAM_METADATA_REGEX)
)


bulk_scrape_mcam_metadata = partial(
    bulk_scrape_metadata, pattern_scraper=scrape_mcam_metadata
)


def parse_mcam_files(file_paths):
    return tuple(
        map(parse_mcam_fn, map(attrgetter("name"), map(Path, file_paths)))
    )


def setup_mcam_bandset_metadata(metadata):
    if "FILTER" in metadata.columns:
        metadata = metadata.drop(
            metadata.loc[metadata.FILTER.isin(("L7", "R7"))].index
        )
        metadata["BAND"] = metadata["FILTER"]
        metadata.drop("FILTER", axis=1)

    metadata.index = metadata["BAND"]
    # add references to the secret bands hidden inside the L0 and R0 images
    bayer_filter_rows = []
    for eye in ("L", "R"):
        if eye + "0" not in metadata.index:
            continue
        eye_row = metadata.loc[eye + "0"]
        for color, ix in zip(("R", "G", "B"), (0, 1, 2)):
            eye_color_row = eye_row.copy()
            eye_color_row["BAND"] = eye + "0" + color
            eye_color_row["IX"] = ix
            eye_color_row.name = eye + "0" + color
            bayer_filter_rows.append(eye_color_row)
        metadata = metadata.drop(eye_row.name)
    if bayer_filter_rows:
        metadata = pd.concat(
            (metadata, pd.concat(bayer_filter_rows, axis=1).T)
        )
    # add wavelengths and bayer pixel mappings
    metadata["WAVELENGTH"] = pd.Series(DERIVED_CAM_DICT["MCAM"]["filters"])[
        metadata["BAND"]
    ]
    metadata["BAYER_PIXEL"] = pd.Series(BAND_TO_BAYER["MCAM"])[
        metadata["BAND"]
    ]
    metadata.loc[metadata["FILETYPE"] != "C00", "BAYER_PIXEL"] = None
    return metadata.reset_index(drop=True)


class McamBandSet(BandSet):
    def __init__(self, observation, rois=None, threads=None):
        files = setup_mcam_bandset_metadata(observation)
        load_method = partial(pdr_load, preserve_constants=[4096, 999999])
        bayer_info = {"pattern": RGGB_PATTERN}
        super().__init__(
            metadata=files,
            load_method=load_method,
            bayer_info=bayer_info,
            rois=rois,
            threads=threads,
        )



from functools import partial

import numpy as np
import pandas as pd
from cytoolz import merge
from numpy.random import default_rng

from marslab.bandset.bandset import log, BandSet

rng = default_rng()


def image_dict_loader(image_dict):
    def load_test_image(band_name):
        return image_dict[band_name]

    return load_test_image


def noise_image_loader(shape):
    def load_noise_image(_):
        return rng.poisson(lam=100, size=shape)

    return load_noise_image


def mock_load(self, generator, *_, **__):
    results = []
    for band_name in self.metadata["BAND"]:
        results.append({band_name: generator(band_name)})
        log.info(f"loaded test {band_name}")
    self.raw = merge(results)


def make_random_bandset():
    num_bands = rng.integers(5, 50)
    shape = rng.integers(1000, 2000, 2)
    band_names = list(range(num_bands))
    bandset = BandSet()
    bandset.metadata = pd.DataFrame()
    bandset.metadata["PATH"] = np.repeat("/fake/path/to/image.png", num_bands)
    bandset.metadata["BAND"] = band_names
    bandset.metadata["IX"] = 0
    bandset.load = partial(mock_load, bandset, noise_image_loader(shape))
    return bandset



"""
constants and functions intended specifically for compatibility with
the behavior of applications outside of this project _other than_
data / metadata format converters.
"""

# TODO: alias mcam, pcam, zcam



from marslab.compat import xcam


def test_derived_cam_dict_construction():
    derived = xcam.DERIVED_CAM_DICT
    assert max(list(derived["ZCAM"]["filters"].values())) == 1022
    assert derived["MCAM"]["virtual_filter_mapping"]["L2_R2"] == ("L2", "R2")
    assert min(list(derived["MCAM"]["filters"].values())) == 445
    assert "L0G_R0G" in derived["ZCAM"]["virtual_filters"]
    assert "L0R_R0R" in derived["MCAM"]["virtual_filters"]



"""
reference implementations of elementary spectrum operations.
overloaded to take dataframes, ndarrays, sequences, and scalars.

generic signature: (reflectance, errors, wavelength). Errors are always
optional. Wavelengths are optional (and ignored) for operations that do not
consider units of wavelength (like ratio), and required for operations that do
(like slope).

order of band shoulders is: left, right, center --
or unordered for things like average

CAUTION: everything in this module must be passed in 'last index fastest' /
'row-primary' orientation. this means that most pandas dataframes need to
be transposed to get meaningful results.
"""
from typing import Optional, Sequence, Union

import numpy as np
from numpy.linalg import norm
import pandas as pd


# TODO: there are a variety of other fairly simple operations we could
#  usefully add to this module.


Specvals = Union[
    pd.DataFrame,
    pd.Series,
    Sequence[float],
    Sequence[Sequence[float]],
    np.ndarray,
]
"""
Valid types for "spectral data" arguments to many functions in this module.
"""


def preprocess_input(
    reflectance: Specvals,
    errors: Optional[Specvals] = None,
    wavelengths: Optional[Specvals] = None,
    band_arity: Optional[int] = None,
    require_wavelengths: bool = False,
) -> tuple[Specvals, Specvals, Specvals]:
    """
    Did someone pass a spectop function something silly? If so, raise an
    exception. Then return all arguments cast to ndarray.
    """
    if reflectance is None:
        raise ValueError("All spectops require reflectance values.")
    if require_wavelengths is True and wavelengths is None:
        raise ValueError("This op requires wavelengths.")
    # this all looks repetitive but the cases are actually different. it is
    # permissible for `errors` and `reflectance` to be 2D arrays, but not
    # for `wavelengths`.
    if hasattr(wavelengths, "ndim") and wavelengths.ndim != 1:
        raise ValueError("Wavelengths must be 1D.")
    # TODO: these should really be None, not array([None]) when not present.
    #  need to change callers to match that signature.
    # noinspection PyTypeChecker
    ref_array, err_array, wave_array = tuple(
        [np.asarray(thing) for thing in (reflectance, errors, wavelengths)]
    )
    if band_arity is not None and ref_array.shape[0] != band_arity:
        raise ValueError(
            f"This op requires {band_arity} bands; got {ref_array.shape[0]}."
        )
    if wavelengths is not None and wave_array.shape[0] != ref_array.shape[0]:
        raise ValueError("There must be as many wavelength values as bands.")
    if errors is not None and err_array.shape != ref_array.shape:
        raise ValueError("Reflectance and errors must have the same shape.")
    return ref_array, err_array, wave_array


# TODO: this smells bad.
def reindex_or_mask(
    *object_pairs: tuple[Optional[Specvals], Optional[Specvals]]
) -> list[Specvals]:
    returning = []
    for pair in object_pairs:
        thing = pair[0]
        if isinstance(pair[1], (pd.DataFrame, pd.Series)):
            thing = pd.Series(thing, index=pair[1].columns)
        elif isinstance(pair[1], Sequence):
            masked = filter(
                lambda a: isinstance(a, np.ma.MaskedArray), pair[1]
            )
            masks = [m.mask for m in masked]
            if len(masks) > 0:
                thing = np.ma.masked_array(thing, mask=sum(masks))
        returning.append(thing)
    return returning


# TODO: we remain uncertain about the validity / relevance of error
#  calculations for some operations.
def addition_in_quadrature(errors: Specvals) -> Optional[Specvals]:
    """
    take what numpy thinks is the norm. In most cases this should be equivalent
    to summing in quadrature. If `errors` contains `None` (generally meaning
    it is an array([None]) produced by `preprocess_input()`), just return None.
    """
    if None in errors:
        return None
    return norm(errors, axis=0)


def ratio(
    reflectance: Specvals,
    errors: Optional[Specvals] = None,
    _wavelengths: Optional[Specvals] = None,
) -> list[Specvals]:
    """
    just a ratio function. notionally, ratio of reflectance values between two
    bands. ignores wavelengths if passed.
    """
    ref_arr, err_arr, _wave_arr = preprocess_input(reflectance, errors, None)
    ratio_value = ref_arr[0] / ref_arr[1]
    if None not in err_arr:
        err_arr = err_arr[0:2]
    err_vals = addition_in_quadrature(err_arr)
    return reindex_or_mask((ratio_value, reflectance), (err_vals, errors))


def band_avg(
    reflectance: Specvals,
    errors: Optional[Specvals] = None,
    _wavelengths: Optional[Specvals] = None,
) -> list[Specvals]:
    """
    Just an averaging function. Notionally, average of all bands within a
    wavelength region. Note that if the caller cares that the bands are
    contiguous wrt an instrument, it is their responsibility to make sure they
    didn't skip any. Ignores `wavelengths` if passed.
    """
    ref_arr, err_arr, _wave_arr = preprocess_input(reflectance, errors, None)
    ref_mean = np.mean(ref_arr, axis=0)
    error_values = addition_in_quadrature(err_arr)
    return reindex_or_mask((ref_mean, reflectance), (error_values, errors))


def slope(
    reflectance: Specvals,
    errors: Optional[Specvals] = None,
    wavelengths: Optional[Specvals] = None,
) -> list[Specvals]:
    """
    Just a slope function. notionally, slope of a line segment in the
    reflectance/wavelength plane drawn between between reflectances and
    center wavelengths oatf two bands.
    """
    ref_arr, err_arr, wave_arr = preprocess_input(
        reflectance, errors, wavelengths, 2, True
    )
    difference, distance = ref_arr[1] - ref_arr[0], wave_arr[1] - wave_arr[0]
    slope_value = difference / distance
    error_values = addition_in_quadrature(err_arr)
    if error_values is not None:
        error_values /= distance
    return reindex_or_mask((slope_value, reflectance), (error_values, errors))


def band_depth(
    reflectance: Specvals,
    errors: Optional[Specvals] = None,
    wavelengths: Specvals = None,
) -> list[Specvals]:
    """
    Standard band depth function. Requires waveleng
    wavelength space between two filters. only looks at the first
    two 'rows' you pass it, should you pass it more.
    """
    ref_arr, err_arr, wave_arr = preprocess_input(
        reflectance, errors, wavelengths, 3, True
    )
    left, right, middle = wave_arr[:3]
    if len({left, right, middle}) != 3:
        raise ValueError(
            "band depth between a wavelength and itself is undefined"
        )
    if not (max(left, right) > middle > min(left, right)):
        raise ValueError(
            "band depth can only be calculated at a band within the "
            "chosen range."
        )
    distance = middle - left
    slope_value = slope(ref_arr[0:2], None, np.array([left, right]))[0]
    continuum_ref = ref_arr[0] + slope_value * distance
    band_depth_value = 1 - ref_arr[2] / continuum_ref
    if (err_vals := addition_in_quadrature(err_arr)) is not None:
        err_vals /= continuum_ref
    return reindex_or_mask((band_depth_value, reflectance), (err_vals, errors))


# TODO, maybe: implement band_depth_min
#  nobody actually seems to want it

# TODO: is there any meaningful way to calculate error
#  for these min and max functions?


def band_min(
    reflectance: Specvals,
    _errors: Optional[Specvals] = None,
    wavelengths: Optional[Specvals] = None,
):
    """
    Returns wavelength at which `reflectance` takes on its minimum value.
    Ignores errors if passed.
    """
    ref_arr, _err_arr, wave_arr = preprocess_input(
        reflectance, None, wavelengths, require_wavelengths=True
    )
    return reindex_or_mask(
        (wave_arr[np.argmin(ref_arr, axis=0)], reflectance), (None, None)
    )


def band_max(
    reflectance: Specvals,
    _errors: Optional[Specvals] = None,
    wavelengths: Optional[Specvals] = None,
) -> list[Specvals]:
    """
    Returns wavelength at which `reflectance` takes on its minimum value.
    Ignores errors if passed.
    """
    ref_arr, _err_arr, wave_arr = preprocess_input(
        reflectance, None, wavelengths, require_wavelengths=True
    )
    return reindex_or_mask(
        (wave_arr[np.argmax(ref_arr, axis=0)], reflectance), (None, None)
    )


SPECTOP_NAMES = (
    "band_avg",
    "band_depth",
    "band_max",
    "band_min",
    "ratio",
    "slope",
)
"""
Names of all functions in this module that follow the Spectop signature 
convention.
"""



import warnings
from itertools import product

import numpy as np

from marslab.imgops import imgutils
from marslab.tests.utilz.div0 import divide_by_zero

RNG = np.random.default_rng()


class TestApplyImageFilter:
    def test_apply_image_filter_1(self):
        out = imgutils.apply_image_filter(
            [[10, 10]],
            image_filter={"function": np.mean, "params": {"axis": 0}},
        )
        assert out[0] == 10
        assert out[1] == 10

    def test_apply_image_filter_2(self):
        out = imgutils.apply_image_filter([[10, 10]])
        assert out[0][0] == 10
        assert out[0][1] == 10


class TestCrop:
    def test_crop_1(self):
        img = np.ones((10, 10))
        out = imgutils.crop(img, bounds=(1, 1, 1, 1))
        assert out.shape[0] == 8
        assert out.shape[1] == 8


class TestEightbit:
    def test_eightbit(self):
        img = np.ones((10, 10))
        img[2] = 0
        out = imgutils.eightbit(img)
        assert out.shape[0] == 10
        assert out.shape[1] == 10
        assert out.dtype == np.uint8
        assert np.all(out[0] == 255)
        assert np.all(out[2] == 0)

    def test_eightbit_dtypes(self):
        """
        Test function for eightbit() dtype handling.
        """
        with warnings.catch_warnings():
            divide_by_zero()
            randarrays = tuple(RNG.random((128, 128)) for _ in range(5))
            dtypes = [*np.typecodes["Float"], *np.typecodes["AllInteger"]]
            for arr, dtype in product(randarrays, dtypes):
                eight = imgutils.eightbit(arr.astype(dtype))
                assert np.isin(eight.min(), (0, 255))
                assert np.isin(eight.max(), (0, 255))
                assert eight.dtype == np.uint8
                assert eight.min() <= eight.max()


def test_normalize_range():
    img = np.ones((10, 10)).astype(np.uint8)
    img[2] = 0
    out = imgutils.normalize_range(img, bounds=(-1, 1))
    assert out[0, 0] == 1
    assert out[2, 0] == -1
    assert out.dtype == np.int16
    fimg = np.random.random((10, 10)).astype(np.float16)
    out32 = imgutils.normalize_range(fimg, bounds=(0, 255))
    assert out32.dtype == np.float32
    assert (out32.min() == 0) and (abs(out32.max() - 255) < 0.01)
    out16 = imgutils.normalize_range(
        fimg, bounds=(0, 255), allow_half_float=True
    )
    assert out16.dtype == np.float16
    assert out16.min() == 0 and out16.max() == 255


def test_normalize_range_masked():
    img = np.ma.masked_array(np.full((10, 10), 2), mask=np.zeros((10, 10)))
    img[2] = 0
    img[3] = 1
    img[5] = 3
    img.mask[5] = True
    out = imgutils.normalize_range(img.astype(np.float32))
    assert out[0, 0] == 1
    assert out[2, 2] == 0
    assert out[3, 3] == 0.5
    assert out.dtype == np.float32


def test_std_clip():
    arr = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    clipped = imgutils.std_clip(arr, 1)
    assert clipped.min() == 2
    assert clipped.max() == 8
    assert clipped.mean() == 5.3
    assert abs(clipped.std() - 2.3259) < 0.01
    assert clipped[4] == arr[4]


def test_cropmask():
    mask = np.ones((10, 10))
    out = imgutils.cropmask(mask, bounds=(1, 1, 1, 1))
    assert out.shape == (10, 10)
    assert np.ma.is_masked(out[0])


def test_crop_all():
    img = np.ones((10, 10))
    imgs = [img.copy() for _ in range(5)]
    out = imgutils.crop_all(imgs, bounds=(1, 1, 1, 1))
    assert out[0].shape[0] == 8
    assert out[1].shape[1] == 8
    assert out[2].shape[0] == 8
    assert out[3].shape[1] == 8
    assert np.all(out[4] == 1)
    assert np.all(out[0] == 1)


def test_clip_unmasked():
    img = np.ma.masked_array(np.ones((10, 10)), mask=np.zeros((10, 10)))
    img[2] = 0
    img[5] = 3
    img.mask[5] = True
    out = imgutils.clip_unmasked(img)
    assert out.min() == 0 and out.max() == 1


def test_bilinear_interpolate_subgrid():
    """
    test function for bilinear_interpolate_subgrid
    """
    test_input = np.array(
        [
            [0, 1, 2, 3],
            [4, 5, 6, 7],
            [8, 9, 10, 11],
            [12, 13, 14, 15],
        ],
        dtype=np.float32,
    )
    rows, columns = np.array([0, 2, 4, 6]), np.array([1, 3, 5, 7])
    result = imgutils.bilinear_interpolate_subgrid(
        rows, columns, test_input, (8, 8)
    )
    expected = np.array(
        [
            [0.0, 0.0, 0.5, 1.0, 1.5, 2.0, 2.5, 3.0],
            [2.0, 2.0, 2.5, 3.0, 3.5, 4.0, 4.5, 5.0],
            [4.0, 4.0, 4.5, 5.0, 5.5, 6.0, 6.5, 7.0],
            [6.0, 6.0, 6.5, 7.0, 7.5, 8.0, 8.5, 9.0],
            [8.0, 8.0, 8.5, 9.0, 9.5, 10.0, 10.5, 11.0],
            [10.0, 10.0, 10.5, 11.0, 11.5, 12.0, 12.5, 13.0],
            [12.0, 12.0, 12.5, 13.0, 13.5, 14.0, 14.5, 15.0],
            [12.0, 12.0, 12.5, 13.0, 13.5, 14.0, 14.5, 15.0],
        ],
        dtype=np.float32,
    )
    assert np.allclose(result, expected)


def test_strict_reshape():
    """
    test function for strict_reshape
    """
    test_input = np.zeros((4, 4))
    test_output = imgutils.strict_reshape(test_input, 1)
    assert test_output.shape == (4, 4)
    test_output = imgutils.strict_reshape(test_input, 3)
    assert test_output.shape == (2, 8)
    test_output = imgutils.strict_reshape(test_input, 0.5)
    assert test_output.shape == (8, 2)



from pathlib import Path

import numpy as np
from PIL import Image
from PIL.JpegImagePlugin import JpegImageFile
from PIL.PngImagePlugin import PngImageFile

from marslab.imgops.imgutils import normalize_range
from marslab.imgops.render import decorrelation_stretch, make_thumbnail
from marslab.tests.utilz.utilz import normal_array

RNG = np.random.default_rng()


def test_decorrelation_stretch_1():
    for _ in range(100):
        channels = [normal_array() for _ in range(3)]
        stretched = decorrelation_stretch(channels, contrast_stretch=1)
        assert np.std(normalize_range(np.dstack(channels))) < np.std(stretched)


def test_make_thumbnail():
    """
    Test the make_thumbnail function.
    """
    try:
        image = RNG.random((1024, 1024, 3))
        jpeg_buf = make_thumbnail(image, (128, 128))
        png_buf = make_thumbnail(image, (512, 512), filetype="png")
        path = make_thumbnail(image, file_or_path_or_buffer="test.png")
        jpeg_buf.seek(0)
        png_buf.seek(0)
        assert JpegImageFile(jpeg_buf).size == (128, 128)
        assert PngImageFile(png_buf).size == (512, 512)
        assert Image.open(path).size == (256, 256)
    finally:
        Path("test.png").unlink(missing_ok=True)



"""
Utilities for watching worker pools. Can be used for progress updates, logging,
etc.

TODO, maybe: this is more special-purpose than originally intended and quite
 possibly shouldn't be in marslab at all.
"""
import os
from collections.abc import Callable, Mapping
import logging
from multiprocessing import active_children, Pool
import time
from types import MappingProxyType
from typing import Optional, TYPE_CHECKING, Union

if TYPE_CHECKING:
    from pathos.multiprocessing import ProcessPool


class ChangeReporter:
    """
    Simple class for monitoring changes in a dict or other mapping.
    """

    def __init__(self, mapping: Mapping):
        self.state = MappingProxyType(mapping)
        self.reference = mapping

    def check(self, new_state: Mapping):
        return {
            key: value
            for key, value in new_state.items()
            if self.state.get(key) != new_state.get(key)
        }

    def update(self, new_state: Mapping):
        changes = self.check(new_state)
        self.state = MappingProxyType(new_state)
        return changes

    def query(self):
        return self.update(self.reference)


def watch_pool(
    result_map: Mapping,
    interval: float = 0.1,
    callback: Optional[Callable] = None,
    timeout: Optional[float] = None,
) -> Mapping:
    start = time.time()
    in_readiness = {
        task: result.ready() for task, result in result_map.items()
    }
    task_report = ChangeReporter(in_readiness)
    while not all(in_readiness.values()):
        if timeout is not None:
            if time.time() - start > timeout:
                raise TimeoutError
        in_readiness = {
            task_ix: result.ready() for task_ix, result in result_map.items()
        }
        report = task_report.update(in_readiness)
        if callback is not None:
            callback(report)
        time.sleep(interval)
    return result_map


def simple_log_callback(
    logger: Optional[logging.Logger] = None, prefix: str = ""
):
    if logger is None:
        logger = logging.getLogger(__name__)

    def log_changed_keys(report):
        for key in report.keys():
            logger.info(prefix + key)

    return log_changed_keys


def wait_for_it(
    pool: Union["ProcessPool", "Pool"],
    results: Mapping,
    log: logging.Logger = None,
    message: str = None,
    callback: Optional[Callable] = None,
    interval: float = 0.1,
    as_dict: bool = False,
    timeout: Optional[float] = None,
) -> Union[dict, list]:
    if (callback is None) and (log is not None):
        callback = simple_log_callback(log, message)
    pool.close()
    if results is not None:
        try:
            watch_pool(results, interval, callback, timeout)
        except TimeoutError:
            pool.terminate()
            for c in active_children():
                os.kill(c.pid, 30)
            raise TimeoutError("Pool timed out")
    pool.join()
    if as_dict:
        return {key: result.get() for key, result in results.items()}
    return [result.get() for result in results.values()]



from pathlib import Path

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from PIL.ImageStat import Stat

from marslab.imgops.loaders import pil_load_rgb
from marslab.imgops.pltutils import get_mpl_image, strip_axes, attach_axis
from marslab.tests.utilz.utilz import render_in_memory
from marslab.tests.data.test_cases import test_data_path

# don't have the tests make a bunch of GUI mess
mpl.use("agg")

# TODO: more cases, clean this up


def test_with_trivial_case():
    red_square = pil_load_rgb(Path(test_data_path, "red_square.png"))
    red_square_array = np.dstack(
        [red_square["R"], red_square["G"], red_square["B"]]
    )
    fig, ax = plt.subplots()
    im = ax.imshow(red_square_array, interpolation="none")
    cax = attach_axis(ax, size="100%", pad="0%")
    colorbar = plt.colorbar(im, cax=cax)
    rendered = render_in_memory(fig)
    # purple-black to green-yellow colorbar, red figure, ticks everywhere
    # a full range of values should be represented in each color band;
    # alpha should be full everywhere (won't be the case if you render it
    # in a Notebook instead!)
    assert rendered.getextrema() == ((0, 255), (0, 255), (0, 255), (255, 255))
    imstats = Stat(rendered)
    # should be redder than it is green
    assert imstats.mean[0] > imstats.mean[1]
    # and also more consistently red
    assert imstats.var[0] < imstats.var[1]
    strip_axes(ax)
    strip_axes(colorbar)
    rendered = render_in_memory(fig)
    # no ticks or spines now. lowest range will still have some red in it.
    assert rendered.getextrema() == ((30, 255), (0, 255), (0, 255), (255, 255))
    # now just get the red square
    final_red_square = get_mpl_image(fig)
    # whose first column should be entirely red
    assert np.all(np.array(final_red_square)[0, :, 0] > 250)



"""
tedious symbolic version of time.py, to use as a basis for more complex
calculations.

attempting to roughly match some outputs of mars24, primarily by reimplementing
algorithms in Allison & McEwen 2000. mostly working.
references:
    1. https://www.giss.nasa.gov/tools/mars24/help/algorithm.html
    2. Allison, M. 1997. Accurate analytic representations of solar time and
        seasons on Mars with applications to the Pathfinder/Surveyor missions.
        Geophys. Res. Lett., 24, 1967-1970.
    3. Allison, M., and M. McEwen 2000. A post-Pathfinder evaluation of
    aerocentric solar coordinates with improved
    timing recipes for Mars seasonal/diurnal climate studies. Planet. Space
    Sci., 48, 215-235. (see /static/docs/Allison_2000.pdf)
"""
import datetime as dt
from typing import Union

import astropy.time as at
import dateutil.parser as dtp
import numpy as np
import sympy as sp


#  ▄         ▄  ▄▄▄▄▄▄▄▄▄▄▄  ▄▄▄▄▄▄▄▄▄▄▄  ▄            ▄▄▄▄▄▄▄▄▄▄▄  ▄▄▄▄▄▄▄▄▄▄▄  ▄▄▄▄▄▄▄▄▄▄▄  ▄▄▄▄▄▄▄▄▄▄▄  ▄▄▄▄▄▄▄▄▄▄▄
# ▐░▌       ▐░▌▐░░░░░░░░░░░▌▐░░░░░░░░░░░▌▐░▌          ▐░░░░░░░░░░░▌▐░░░░░░░░░░░▌▐░░░░░░░░░░░▌▐░░░░░░░░░░░▌▐░░░░░░░░░░░▌
# ▐░▌       ▐░▌ ▀▀▀▀█░█▀▀▀▀  ▀▀▀▀█░█▀▀▀▀ ▐░▌           ▀▀▀▀█░█▀▀▀▀  ▀▀▀▀█░█▀▀▀▀  ▀▀▀▀█░█▀▀▀▀ ▐░█▀▀▀▀▀▀▀▀▀ ▐░█▀▀▀▀▀▀▀▀▀
# ▐░▌       ▐░▌     ▐░▌          ▐░▌     ▐░▌               ▐░▌          ▐░▌          ▐░▌     ▐░▌          ▐░▌
# ▐░▌       ▐░▌     ▐░▌          ▐░▌     ▐░▌               ▐░▌          ▐░▌          ▐░▌     ▐░█▄▄▄▄▄▄▄▄▄ ▐░█▄▄▄▄▄▄▄▄▄
# ▐░▌       ▐░▌     ▐░▌          ▐░▌     ▐░▌               ▐░▌          ▐░▌          ▐░▌     ▐░░░░░░░░░░░▌▐░░░░░░░░░░░▌
# ▐░▌       ▐░▌     ▐░▌          ▐░▌     ▐░▌               ▐░▌          ▐░▌          ▐░▌     ▐░█▀▀▀▀▀▀▀▀▀  ▀▀▀▀▀▀▀▀▀█░▌
# ▐░▌       ▐░▌     ▐░▌          ▐░▌     ▐░▌               ▐░▌          ▐░▌          ▐░▌     ▐░▌                    ▐░▌
# ▐░█▄▄▄▄▄▄▄█░▌     ▐░▌      ▄▄▄▄█░█▄▄▄▄ ▐░█▄▄▄▄▄▄▄▄▄  ▄▄▄▄█░█▄▄▄▄      ▐░▌      ▄▄▄▄█░█▄▄▄▄ ▐░█▄▄▄▄▄▄▄▄▄  ▄▄▄▄▄▄▄▄▄█░▌
# ▐░░░░░░░░░░░▌     ▐░▌     ▐░░░░░░░░░░░▌▐░░░░░░░░░░░▌▐░░░░░░░░░░░▌     ▐░▌     ▐░░░░░░░░░░░▌▐░░░░░░░░░░░▌▐░░░░░░░░░░░▌
#  ▀▀▀▀▀▀▀▀▀▀▀       ▀       ▀▀▀▀▀▀▀▀▀▀▀  ▀▀▀▀▀▀▀▀▀▀▀  ▀▀▀▀▀▀▀▀▀▀▀       ▀       ▀▀▀▀▀▀▀▀▀▀▀  ▀▀▀▀▀▀▀▀▀▀▀  ▀▀▀▀▀▀▀▀▀▀▀


def parse_time(time: Union[dt.datetime, at.Time, str, float]) -> at.Time:
    """
    attempts to coerce any input to an astropy.Time object.
    assumes float inputs are in julian days.
    """
    if isinstance(time, (dt.datetime, at.Time)):
        return at.Time(time)
    elif isinstance(time, float):
        return at.Time(time, format="jd")
    return at.Time(dtp.parse(time))


def to_jd_tt(time: Union[dt.datetime, at.Time, str, float, None]) -> float:
    if time is None:
        time = at.Time.now()
    return parse_time(time).tt.jd


def to_j2000_jd(time: Union[dt.datetime, at.Time, str, float]) -> float:
    """
    time difference between time and J2000 epoch in julian days,
    and specifically terrestrial time
    this value (delta_j2000 and variants below) is a basic
    time unit for many common expressions in Mars horology
    """
    time = parse_time(time)
    return (at.Time(time) - at.Time("2000-01-01T12:00:00")).tt.jd


def d2r(degree_value: float) -> float:
    """degrees to radians"""
    return degree_value * np.pi / 180


def r2d(radian_value: float) -> float:
    """radians to degrees"""
    return radian_value * 180 / np.pi


def hours_to_24h_time(hours: float) -> str:
    """
    convoluted expression because of Python's...thing to transform a figure
    given in fractional hours to a string in ISO time format
    """
    return (
        (dt.datetime(2001, 1, 2) + dt.timedelta(hours=hours))
        .time()
        .isoformat()
    )


#  ▄▄       ▄▄  ▄▄▄▄▄▄▄▄▄▄▄  ▄▄▄▄▄▄▄▄▄▄▄  ▄         ▄  ▄▄▄▄▄▄▄▄▄▄▄
# ▐░░▌     ▐░░▌▐░░░░░░░░░░░▌▐░░░░░░░░░░░▌▐░▌       ▐░▌▐░░░░░░░░░░░▌
# ▐░▌░▌   ▐░▐░▌▐░█▀▀▀▀▀▀▀█░▌ ▀▀▀▀█░█▀▀▀▀ ▐░▌       ▐░▌▐░█▀▀▀▀▀▀▀▀▀
# ▐░▌▐░▌ ▐░▌▐░▌▐░▌       ▐░▌     ▐░▌     ▐░▌       ▐░▌▐░▌
# ▐░▌ ▐░▐░▌ ▐░▌▐░█▄▄▄▄▄▄▄█░▌     ▐░▌     ▐░█▄▄▄▄▄▄▄█░▌▐░█▄▄▄▄▄▄▄▄▄
# ▐░▌  ▐░▌  ▐░▌▐░░░░░░░░░░░▌     ▐░▌     ▐░░░░░░░░░░░▌▐░░░░░░░░░░░▌
# ▐░▌   ▀   ▐░▌▐░█▀▀▀▀▀▀▀█░▌     ▐░▌     ▐░█▀▀▀▀▀▀▀█░▌ ▀▀▀▀▀▀▀▀▀█░▌
# ▐░▌       ▐░▌▐░▌       ▐░▌     ▐░▌     ▐░▌       ▐░▌          ▐░▌
# ▐░▌       ▐░▌▐░▌       ▐░▌     ▐░▌     ▐░▌       ▐░▌ ▄▄▄▄▄▄▄▄▄█░▌
# ▐░▌       ▐░▌▐░▌       ▐░▌     ▐░▌     ▐░▌       ▐░▌▐░░░░░░░░░░░▌
#  ▀         ▀  ▀         ▀       ▀       ▀         ▀  ▀▀▀▀▀▀▀▀▀▀▀

# all equation / table numbers in this section of the code
# refer to Allison & McEwen 2000

# determine mars mean anomaly, fictitious mean sun angle,
# and sum of angular perturbations in longitude

# TODO: it's not incredibly useful as it stands to define these
#  as sympy expressions, but i'm anticipating further applications
#  where it may be. I don't think it's significantly hurting
#  performance _other than imports_ because we immediately lambdify
#  them for evaluation. if we are building a quick command-line utility
#  and it appears to be discernibly laggy, or people want to run these on
#  > 1000 data points at once, we may want to try slimming this down.

# this is a free symbol that will generally represent an
# output of j2000_jd
delta_j2000 = sp.S("Delta_j2000")

# eq. 16
# note that all angular values in A&M are
# in degrees; we convert them to radians
mean_anomaly = d2r(19.3871) + d2r(0.52402073) * delta_j2000

# eq 17
fictitious_mean_sun = d2r(270.3871) + d2r(0.524038496) * delta_j2000

# table 5 (truncated)
perturbation_table = np.array(
    [
        [d2r(0.0071), 2.2353, d2r(49.409)],
        [d2r(0.0057), 2.7543, d2r(168.173)],
        [d2r(0.0039), 1.1177, d2r(191.837)],
        [d2r(0.0037), 15.7866, d2r(21.736)],
        [d2r(0.0021), 2.1354, d2r(15.704)],
        [d2r(0.0020), 2.4694, d2r(95.528)],
        [d2r(0.0018), 32.8493, d2r(49.095)],
    ]
)

# free symbols representing terms of perturbation function
# / columns of perturbation table
A, tau, phi = sp.S("A, tau, phi")

# eq 18, interior of summation
PBS_i = A * sp.cos(d2r(0.985626) * delta_j2000 / tau + phi)


# evaluation function for this sum
def pbs(dt_j2000: float) -> np.ndarray:
    row_lambda = sp.lambdify([delta_j2000, A, tau, phi], PBS_i)
    return np.sum(
        [
            row_lambda(dt_j2000, *perturbation_table[i])
            for i in range(perturbation_table.shape[0])
        ]
    )


# terms of equation 19:
mean_terms = (
    (d2r(10.691) + d2r(3e-7) * delta_j2000) * sp.sin(mean_anomaly)
    + d2r(0.623) * sp.sin(2 * mean_anomaly)
    + d2r(0.050) * sp.sin(3 * mean_anomaly)
    + d2r(0.005) * sp.sin(4 * mean_anomaly)
    + d2r(0.0005) * sp.sin(5 * mean_anomaly)
)


# evaluation function: areocentric solar longitude
def l_s(dt_j2000: float) -> float:
    return (
        sp.lambdify(delta_j2000, fictitious_mean_sun)(dt_j2000)
        + sp.lambdify(delta_j2000, mean_terms)(dt_j2000)
        + pbs(dt_j2000)
    )


# note that areocentric solar longitude is given in Mars24;
# can compare, for instance:
# r2d(L_s(j2000_jd(at.Time.now()))) % 360

# eq 20
def eot(dt_j2000: float) -> float:
    """Allison's 'equation of time'"""
    # just expressing this one numerically for now
    return (
        d2r(2.861) * np.sin(2 * l_s(dt_j2000))
        - d2r(0.071) * np.sin(4 * l_s(dt_j2000))
        + d2r(0.002) * np.sin(6 * l_s(dt_j2000))
        + sp.lambdify(delta_j2000, fictitious_mean_sun)(dt_j2000)
        - l_s(dt_j2000)
    )
    # these final two terms are -(v-m), additive inverse
    # of equation of center


# ******* cutting to the chase, the simplistic useful expressions *******

# NOTE 1: we _stop_ modifying to radians here, because these will be public
# quite shortly and aren't using any more transcendental functions
# NOTE 2: following Allison & McEwan, we _also_ stop using delta_j2000
# and switch to conventionally-defined jd: but still in TT!

# eq 22, modified constants (see giss.nasa.gov link, heading C-2)
def mst_hours(julian_day: float) -> float:
    days = (julian_day - 2451549.5) / 1.0274912517 + 44796.0 - 0.0009626
    return (days * 24) % 24


# eq. 23
def lmst_hours(julian_day: float, west_longitude_degrees: float) -> float:
    return mst_hours(julian_day) - west_longitude_degrees * 24 / 360


def ltst_hours(julian_day: float, west_longitude_degrees: float) -> float:
    hours = (
        lmst_hours(julian_day, west_longitude_degrees)
        + r2d(eot((to_j2000_jd(at.Time(julian_day, format="jd"))))) / 15
    )
    return hours


#  ▄▄▄▄▄▄▄▄▄▄▄  ▄         ▄  ▄▄▄▄▄▄▄▄▄▄▄  ▄▄▄▄▄▄▄▄▄▄▄  ▄         ▄  ▄▄▄▄▄▄▄▄▄▄▄
# ▐░░░░░░░░░░░▌▐░▌       ▐░▌▐░░░░░░░░░░░▌▐░░░░░░░░░░░▌▐░▌       ▐░▌▐░░░░░░░░░░░▌
# ▐░█▀▀▀▀▀▀▀█░▌▐░▌       ▐░▌ ▀▀▀▀█░█▀▀▀▀ ▐░█▀▀▀▀▀▀▀█░▌▐░▌       ▐░▌ ▀▀▀▀█░█▀▀▀▀
# ▐░▌       ▐░▌▐░▌       ▐░▌     ▐░▌     ▐░▌       ▐░▌▐░▌       ▐░▌     ▐░▌
# ▐░▌       ▐░▌▐░▌       ▐░▌     ▐░▌     ▐░█▄▄▄▄▄▄▄█░▌▐░▌       ▐░▌     ▐░▌
# ▐░▌       ▐░▌▐░▌       ▐░▌     ▐░▌     ▐░░░░░░░░░░░▌▐░▌       ▐░▌     ▐░▌
# ▐░▌       ▐░▌▐░▌       ▐░▌     ▐░▌     ▐░█▀▀▀▀▀▀▀▀▀ ▐░▌       ▐░▌     ▐░▌
# ▐░▌       ▐░▌▐░▌       ▐░▌     ▐░▌     ▐░▌          ▐░▌       ▐░▌     ▐░▌
# ▐░█▄▄▄▄▄▄▄█░▌▐░█▄▄▄▄▄▄▄█░▌     ▐░▌     ▐░▌          ▐░█▄▄▄▄▄▄▄█░▌     ▐░▌
# ▐░░░░░░░░░░░▌▐░░░░░░░░░░░▌     ▐░▌     ▐░▌          ▐░░░░░░░░░░░▌     ▐░▌
#  ▀▀▀▀▀▀▀▀▀▀▀  ▀▀▀▀▀▀▀▀▀▀▀       ▀       ▀            ▀▀▀▀▀▀▀▀▀▀▀       ▀

# end-user wrapper functions


def mst(time: Union[at.Time, float, str, None, dt.datetime] = None) -> str:
    jd_tt = to_jd_tt(time)
    return hours_to_24h_time(mst_hours(jd_tt))


def lmst(
    time: Union[at.Time, float, str, None, dt.datetime] = None,
    west_longitude: float = 0,
) -> str:
    jd_tt = to_jd_tt(time)
    return hours_to_24h_time(lmst_hours(jd_tt, west_longitude))


def ltst(
    time: Union[at.Time, float, str, None, dt.datetime] = None,
    west_longitude: float = 0,
) -> str:
    jd_tt = to_jd_tt(time)
    return hours_to_24h_time(ltst_hours(jd_tt, west_longitude))



from functools import reduce

import numpy as np


def dcs(channels):
    input_shape = (*channels[0].shape, 3)
    channel_vectors = np.dstack(channels).reshape(-1, 3)
    channel_covariance = np.ma.cov(channel_vectors.T)
    eigenvalues, eigenvectors = np.linalg.eig(channel_covariance)
    # diagonal matrix containing per-band "stretch factors"
    stretch_matrix = np.diag(1 / np.sqrt(eigenvalues))
    # mean values for each channel
    channel_means = np.mean(channel_vectors, axis=0)
    # full transformation matrix:
    # rotates into eigenspace of covariance matrix, applies stretch,
    # rotates back to channelspace, applies sigma scaling
    transformation_matrix = reduce(
        np.dot, [eigenvectors, stretch_matrix, eigenvectors.T]
    )

    # remove mean from each channel, transform, replace mean and add offset
    dcs_vectors = (
        np.dot((channel_vectors - channel_means), transformation_matrix)
        + channel_means
    )
    dcs_array = dcs_vectors.reshape(input_shape)
    # special limiter included ensuite
    for i in range(3):
        c = dcs_array[:, :, i]
        dcs_array[:, :, i] -= c.min()
        dcs_array[:, :, i] *= 1/c.max()
    return dcs_array



from io import BytesIO

from hypothesis import strategies as st
from matplotlib.figure import Figure
import numpy as np
from PIL import Image


def boring_reals(min_mag: int, max_mag: int):
    """
    hypothesis strategy that returns finite floats within a specified range of
    absolute magnitudes.
    """
    return st.floats(
        allow_infinity=False,
        allow_nan=False,
        allow_subnormal=False,
        min_value=-(10 ** max_mag),
        max_value=10 ** max_mag,
    ).filter(lambda x: x == 0 or min_mag <= x <= max_mag)


quaternions = st.lists(boring_reals(-4, 8), min_size=4, max_size=4)
finite_floats = st.floats(allow_nan=False, allow_infinity=False)
finite_numbers = st.one_of(st.integers(), finite_floats)
positive_finite_floats = st.floats(
    min_value=0, allow_nan=False, allow_infinity=False
)


def normal_array(loc=128, scale=16, size=(256, 256), dtype=np.int64):
    rng = np.random.default_rng()
    randarray = rng.normal(loc, scale, size)
    return randarray.astype(dtype)


def render_in_memory(fig: Figure) -> Image:
    outbuf = BytesIO()
    fig.savefig(outbuf)
    outbuf.seek(0)
    return Image.open(outbuf)



from marslab.imgops.look import Look


def test_look():
    """
    Very basic test for Look.
    """
    from matplotlib.figure import Figure
    import numpy as np

    from marslab.imgops import render
    from marslab.imgops import masking
    from marslab.imgops import imgutils

    def enhancer(arrs, mask):
        masked = np.ma.masked_array(
            np.dstack(arrs), mask=np.tile(mask, (1, 1, 3))
        )
        return imgutils.enhance_color(masked, (0, 1), 1)

    list_ = []
    instruction = {
        "crop": (1, 1, 1, 1),
        "prefilter": {
            "function": imgutils.normalize_range,
            "params": {"stretch": 2},
        },
        "look": enhancer,
        "mask": {
            "instructions": [
                {
                    "function": masking.threshold_mask,
                    "params": {"percentiles": (55, 95)},
                    "colorfill": {"color": 0, "mask_alpha": 1},
                    "pass": True,
                    "send": True,
                }
            ]
        },
        "plotter": {"function": render.simple_figure},
        "bang": {"function": lambda: list_.append(1)},
    }
    look = Look.compile_from_instruction(instruction)
    r = np.random.random((128, 128))
    g = np.random.random((128, 128)) * 10
    b = np.random.random((128, 128)) * 100
    fig = look.execute((r, g, b))
    assert isinstance(fig, Figure)
    assert look.inserts["plotter"]["layers"][0][0].shape == (126, 126, 4)
    assert list_ == [1]



"""
inline rendering functions for look pipelines. can also be called on their own.
"""
from __future__ import annotations

import io
from functools import reduce
from itertools import chain, repeat
from pathlib import Path
from typing import (
    Callable,
    Collection,
    Mapping,
    Optional,
    Sequence,
    TypedDict,
    TYPE_CHECKING,
    Union,
)

from dustgoggles.structures import separate_by
import matplotlib as mpl
import matplotlib.cm as cm
import matplotlib.figure
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.font_manager import FontProperties
from PIL import Image

from marslab.imgops.debayer import make_bayer, debayer_upsample
from marslab.imgops.imgutils import (
    colorfill_maskedarray,
    eightbit,
    enhance_color,
    normalize_range,
)
from marslab.imgops.pltutils import (
    attach_axis,
    get_mpl_image,
    set_colorbar_font,
    strip_axes,
)
from marslab.spectops import Specvals

if TYPE_CHECKING:
    from marslab.imgops.look import LookInstruction


def decorrelation_stretch(
    channels: Sequence[np.ndarray],
    flat_mask: Optional[np.ndarray] = None,
    contrast_stretch: Optional[Union[Sequence[float], float]] = None,
    sigma: Optional[float] = 1,
    mask_fill_tone: Optional[Union[float, tuple[float, float, float]]] = None,
) -> np.ndarray:
    """
    decorrelation stretch of passed array on last axis of array. see
    Gillespie et al. 1986, etc., etc.

    This is partly an adaptation of the MATLAB DCS implementation.
    Work towards this adaptation is partly due to Lukáš Brabec
    (github.com/lbrabec/decorrstretch) and Christian Tate (unreleased).

    channels: must be at least two ndarrays, numeric dtypes. each ndarray
    should be 2D (size > 1). no infs or nans. no channel may have entirely
    equal elements. extremely large variation in array means between channels
    relative to data type may result in errors.
    integer arrays will be cast to float32. float arrays of depth < 32 will
    be cast to float32.
    contrast_stretch: float or sequence of two floats. each must be
    between 0 and 100 inclusive. percentile ranges to use as max/min bounds
    of output array.
    sigma: fixed target standard deviation for each channel. must be >= 0;
    ranges between 0 and 1 recommended. None means that the original standard
    deviation per channel is used as the target (this is a 'classic'
    decorrelation stretch). Doesn't matter if you're applying a contrast
    stretch.
    """
    if any(isinstance(c, np.ma.MaskedArray) for c in channels):
        working_array = np.ma.dstack(channels)
    else:
        working_array = np.dstack(channels)
    if flat_mask is not None:
        flat_mask = np.dstack([flat_mask] * 3)
        if not isinstance(working_array, np.ma.MaskedArray):
            working_array = np.ma.masked_array(working_array, mask=flat_mask)
        else:
            working_array.mask += flat_mask
    input_shape = working_array.shape
    channel_vectors = working_array.reshape(-1, input_shape[-1])
    if channel_vectors.dtype not in [np.float32, np.float64]:
        working_dtype = np.float32
    else:
        working_dtype = channel_vectors.dtype
    if (
        isinstance(channel_vectors, np.ma.MaskedArray)
        and (cmask := channel_vectors.mask).size > 1
    ):
        selector = cmask[:, 0] | cmask[:, 1] | cmask[:, 2]
        valid_vectors = channel_vectors[~selector]
    else:
        valid_vectors = channel_vectors
    channel_covariance = np.ma.cov(valid_vectors.T).astype(working_dtype)
    eigenvalues, eigenvectors = np.linalg.eig(channel_covariance)
    # diagonal matrix containing per-band "stretch factors"
    stretch_matrix = np.diag(1 / np.sqrt(eigenvalues))
    # mean values for each channel
    channel_means = np.mean(valid_vectors, axis=0, dtype=working_dtype)
    # full transformation matrix:
    # rotates into eigenspace of covariance matrix, applies stretch,
    # rotates back to channelspace, applies sigma scaling
    transformation_matrix = reduce(
        np.dot, [eigenvectors, stretch_matrix, eigenvectors.T]
    )
    # target per-channel standard deviation as a diagonalized matrix.
    # set equal to sigma if sigma is passed; otherwise simply set
    # equal to per-channel input standard deviation. if it's 1, skip the step.
    if sigma is None:
        channel_sigmas = np.diag(np.sqrt(channel_covariance.diagonal()))
        transformation_matrix = np.dot(transformation_matrix, channel_sigmas)
    elif sigma != 1:
        channel_sigmas = np.diag(
            np.array([sigma for _ in range(len(channels))])
        )
        transformation_matrix = np.dot(transformation_matrix, channel_sigmas)
    # TODO: this 'offset' term does not explicitly exist in the matlab
    #  implementation. check against reference algorithm. it rarely does
    #  anything, though.
    offset = channel_means - np.dot(channel_means, transformation_matrix)
    # remove mean from each channel, transform, replace mean and add offset
    dcs_vectors = (
        np.dot((channel_vectors - channel_means), transformation_matrix)
        + channel_means
        + offset
    )
    dcs_array = dcs_vectors.reshape(input_shape)
    # special limiter included ensuite
    if contrast_stretch is None:
        image = normalize_range(dcs_array)
    else:
        image = enhance_color(dcs_array, (0, 1), contrast_stretch)
    if (isinstance(image, np.ma.MaskedArray)) and (mask_fill_tone is not None):
        image = colorfill_maskedarray(image, mask_fill_tone)
    return image


# TODO: I think masking in this is too late and we should be doing it up
#  front, adding an optional median fill step for cases where we might have
#  undesirable white pixels everywhere or whatever.
def render_rgb_composite(
    channels: Sequence[np.ndarray],
    *,
    special_constants: Optional[Union[Collection[float], np.ndarray]] = None,
) -> np.ndarray:
    """
    Composite three input arrays / 'channels' into an 'image', All three arrays
    must have the same shape, and for best results, should be of the same
    dtype. The returned 'image' has shape (*input_shape, 3).

    Although not limited to this application, this function is intended
    primarily to merge 2D arrays representing red, green, and blue image
    channels into an array that can be easily rendered as an RGB image.

    This function is a good basis for producing both "true-color" and
    "enhanced-color" images from most filter sets, and can also be used for
    stranger purposes.

    TODO: following is no longer true. move this as an interpretation step to
      Look.compile_from_instruction()
    this assumes normalization as a default option because you're presumably
    going to want to view these as "normal" RGB images, not scaled to an
    arbitrary colormap (although aren't they all?).
    """
    assert len(channels) == 3
    assert len(set(map(lambda arr: arr.shape, channels))) == 1
    if isinstance(channels[0], np.ma.MaskedArray):
        composed = np.ma.dstack(channels)
    else:
        composed = np.dstack(channels)
    if special_constants is None:
        return composed
    special_mask = np.isin(composed, special_constants)
    if isinstance(composed, np.ma.MaskedArray):
        composed.mask = np.logical_or(composed.mask, special_mask)
    else:
        composed = np.ma.MaskedArray(composed, special_mask)
    return composed


def spectop_look(
    images: Union[Specvals],
    flat_mask: Optional[np.ndarray] = None,
    *,
    spectop: Callable[
        [Specvals, Optional[Specvals], Optional[Specvals]],
        tuple[np.ndarray, np.ndarray],
    ],
    wavelengths: Optional[Specvals] = None,
):
    look = spectop(images, None, wavelengths)[0]
    if flat_mask is None:
        return look
    if isinstance(look, np.ma.MaskedArray):
        look.mask += flat_mask
    else:
        look = np.ma.MaskedArray(look, mask=flat_mask)
    return look


# TODO: cruft, this should be handled by BandSet -- but is it?
def rgb_from_bayer(
    image: np.ndarray,
    bayer_pattern: Mapping[str, Sequence[int]],
    bayer_pixels: Sequence[Union[str, tuple[str]]] = (
        "red",
        ("green_1", "green_2"),
        "blue",
    ),
) -> np.ndarray:
    """
    assemble m x n x 3 array from specified bayer pixels of passed
    single m x n array or 3 m x n arrays
    """
    assert len(bayer_pixels) == 3, "this function only makes 3-channel images"
    if not isinstance(image, Sequence):
        image = repeat(image)
    elif len(image) == 1:
        image = repeat(image[0])
    else:
        assert (
            len(image) == 3
        ), "this function operates on either one or three input arrays"
    channels = []
    bayer_masks = None
    bayer_row_column = None
    for image, bayer_pixel in zip(image, bayer_pixels):
        if bayer_masks is None:  # i.e., the first time
            bayer_masks = make_bayer(image.shape, bayer_pattern)
            bayer_row_column = {
                pixel: (np.unique(mask[0]), np.unique(mask[1]))
                for pixel, mask in bayer_masks.items()
            }
        channels.append(
            debayer_upsample(
                image,
                bayer_pixel,
                masks=bayer_masks,
                row_column=bayer_row_column,
            )
        )
    return np.dstack(channels)


def make_thumbnail(
    image_array: np.ndarray,
    thumbnail_size: tuple[int, int] = (256, 256),
    file_or_path_or_buffer: Optional[Union[str, Path, io.BytesIO]] = None,
    filetype: Optional[str] = None,
) -> Union[str, Path, io.BytesIO]:
    """
    makes thumbnails from arrays or matplotlib images or PIL.Images
    """
    if filetype is None and not isinstance(
        file_or_path_or_buffer, (str, Path)
    ):
        filetype = "jpeg"
    if file_or_path_or_buffer is None:
        file_or_path_or_buffer = io.BytesIO()
    if isinstance(image_array, mpl.figure.Figure):
        thumbnail_array = get_mpl_image(image_array).convert("RGB")
    elif isinstance(image_array, np.ndarray):
        image_array = eightbit(image_array)
        thumbnail_array = Image.fromarray(image_array).convert("RGB")
    else:
        thumbnail_array = image_array
    thumbnail_array.thumbnail(thumbnail_size)
    thumbnail_array.save(file_or_path_or_buffer, filetype)
    return file_or_path_or_buffer


# TODO: probably all of this special-case special constants handling is
#  becoming hacky -- maybe solve it more consistently with masks?:
def colormapped_plot(
    array: np.ndarray,
    cmap: Union[str, mpl.colors.Colormap, None] = None,
    render_colorbar: bool = False,
    no_ticks: bool = True,
    colorbar_fp: Optional[FontProperties] = None,
    special_constants: Optional[Sequence[Union[int, float]]] = None,
    mask_fill_color: Union[float, tuple[float, float, float]] = 0.45,
    drop_mask: bool = True,
    alpha: Optional[float] = None,
    layers: Optional[
        Union[np.ndarray, dict[str, Union[int, np.ndarray]]]
    ] = None,
    n_ticks=6,
):
    """
    generate a colormapped plot, optionally with colorbar, from 2D array.
    render as many overlays and underlays as you like.
    layers may either be ndarrays, dicts of the format:
    {'layer_ix': int, 'image': np.ndarray},
    (where the colormapped array is always at layer_ix 0),
    or lists or tuples of the same.
    ndarrays without a layer_ix field will always be treated as overlays,
    rendered in argument order after the layer with the highest specified
    layer_ix.
    3D layers will be treated as RGB(A); 2D layers will be treated as
    grayscale.
    Note that if opacity is not set and cmap doesn't define any
    transparency, you won't see any underlays.
    """
    # TODO: hacky bailout if this is stuck on the end of an overlay pipeline,
    #  this can be cleaned up much more effectively
    if isinstance(array, mpl.figure.Figure):
        return array
    normalization_array = array.copy()
    if isinstance(array, np.ma.masked_array):
        normalization_array = normalization_array[~normalization_array.mask]
    if special_constants is not None:
        normalization_array = normalization_array[
            ~np.isin(normalization_array, special_constants)
        ]
    normalization_array = normalization_array[np.isfinite(normalization_array)]
    # this now should be using masks appropriately...but perhaps it is not
    norm = plt.Normalize(
        vmin=normalization_array.min(), vmax=normalization_array.max()
    )
    del normalization_array
    if isinstance(array, np.ma.masked_array):
        if drop_mask is True:
            array = array.data
    if isinstance(cmap, str):
        cmap = plt.get_cmap(cmap)
    elif cmap is None:
        cmap = plt.get_cmap("Greys_r")
    elif not isinstance(cmap, plt.Colormap):
        raise TypeError(f"`cmap` must be NoneType, str, or Colormap")
    mapped = cmap(norm(array))
    if isinstance(array, np.ma.MaskedArray) and (drop_mask is False):
        # noinspection PyTupleItemAssignment,PyUnresolvedReferences
        mapped[array.mask] = mask_fill_color
    del array
    if alpha is not None:
        mapped[:, :, 3] = alpha
    if layers is not None:
        fig, ax = flatten_into_figure(
            list(layers) + [{"layer_ix": 0, "image": mapped}]
        )
    else:
        fig, ax = plt.subplots()
        ax.imshow(mapped)
    if render_colorbar:
        attach_colorbar(ax, cmap, colorbar_fp, norm, n_ticks)
    if no_ticks:
        strip_axes(ax)
    return fig


def _tformat(number: float, order: int, precision: int):
    """Standard compact tick label rounder / formatter."""
    if order > 2:
        # use exponential notation when value range is < 0.1
        formatted = ("{:." + str(precision) + "e}").format(number)
        # remove pointless leading 0 in exponent
        return "".join([formatted[:-2], formatted[-1]])
    if order + precision < 0:
        return str(round(number))
    return str(number)


# TODO, maybe: this only works well with certain sorts of value ranges.
def _trylabel(
    bounds: tuple[float, float], order: int, precision: int, n_ticks: int
) -> tuple[Union[np.ndarray, list[str]], bool]:
    """
    Attempts to find label positions that maximize coverage while minimizing
    required decimal places.
    """
    # 100 is arbitrary, basically the resolution of the solver
    grid = np.linspace(bounds[0], bounds[1], n_ticks * 100)
    labels = sorted(
        {_tformat(round(n, order + precision), order, precision) for n in grid}
    )
    labels = np.array(labels)
    positions = labels.astype(float)
    inbounds = (positions <= grid.max()) & (positions >= grid.min())
    labels, positions = labels[inbounds], positions[inbounds]
    labels, positions = labels[np.argsort(positions)], np.sort(positions)
    # NOTE: 20% range cutoff here is totally arbitrary
    if (
        positions.size < n_ticks
        or abs((positions.max() - grid.max()) / grid.max()) > 0.2
        or abs((positions.min() - grid.min()) / grid.min()) > 0.2
    ):
        # noinspection PyTypeChecker
        return labels, False
    outlabels = []
    for ideal in np.linspace(bounds[0], bounds[1], n_ticks):
        outlabels.append(labels[np.argmin(abs(positions - ideal))])
    # noinspection PyTypeChecker
    return outlabels, True


def attach_colorbar(
    ax: plt.Axes,
    cmap: Union[str, mpl.colors.Colormap],
    colorbar_fp: Optional[mpl.font_manager.FontProperties] = None,
    norm: Optional[mpl.colors.Normalize] = None,
    n_ticks: int = 5,
):
    """
    Attach a colorbar tightly to a matplotlib axis. Is typically much better
    aligned than default colorbar settings.
    """
    cax = attach_axis(ax, size="3%", pad="0.5%")
    colorbar = plt.colorbar(cm.ScalarMappable(norm=norm, cmap=cmap), cax=cax)
    ymin, ymax = colorbar.ax.get_ylim()
    extent = ymax - ymin
    order = int(1 - np.floor(np.log10(extent)))
    precision, success = 0, False
    while success is False:
        labels, success = _trylabel((ymin, ymax), order, precision, n_ticks)
        precision += 1
        if precision > 10:
            raise ValueError("Tick placement not terminating; check range.")
    # noinspection PyTypeChecker
    # noinspection PyUnboundLocalVariable
    colorbar.ax.set_yticks(list(map(float, labels)))
    colorbar.ax.set_yticklabels(labels)
    if colorbar_fp:
        set_colorbar_font(colorbar, colorbar_fp)


def _duck_alpha(rgb: np.ndarray, a: np.ndarray) -> np.ndarray:
    """
    Performs elementwise multiplication (along axes 0 and 1) between a 3-D
    and 2-D array. Used to apply an alpha channel when merging layers.
    """
    return np.einsum("ijk,ij->ijk", rgb, a)


def merge_layers(lower: np.ndarray, upper: np.ndarray) -> np.ndarray:
    """
    "Flatten" two arrays representing image layers. Each array must be 3D,
    and each must have shape (n, m, 3) or (n, m, 4). If the last axis is of
    length 4, the last "band" of the array is interpreted as an alpha channel.
    Returns an array of shape (n, m, 4) unless `upper` is of shape (n, m, 3),
    in which case it just returns `upper` -- it has no transparency so fully
    covers up `lower` in the flattened image.
    """
    if upper.shape[2] == 3:
        return upper
    alpha_upper = upper[:, :, 3]
    if lower.shape[2] == 3:
        alpha_lower = 1 - alpha_upper
    else:
        alpha_lower = np.min([1 - alpha_upper, lower[:, :, 3]], axis=0)
    rgb = _duck_alpha(upper[:, :, :3], alpha_upper) + _duck_alpha(
        lower[:, :, :3], alpha_lower
    )
    return np.dstack([rgb, alpha_upper + alpha_lower])


class LayerSpec(TypedDict):
    """
    Format for a dict to pass to layer-merging functions if you want to ensure
    an array goes at a specific place in the stack.
    """

    image: np.ndarray
    layer_ix: Optional[int]


def flatten_into_figure(
    layers: Sequence[Union[LayerSpec, np.ndarray]], **imshow_kwargs
) -> tuple[plt.Figure, plt.Axes]:
    seq, single = separate_by(layers, lambda l: isinstance(l, (tuple, list)))
    layers = list(chain.from_iterable(seq)) + single
    dicts, arrays = separate_by(layers, lambda l: isinstance(l, dict))
    order = [d.get("layer_ix") for d in dicts]
    assert len(set(order)) == len(order)
    dicts.sort(key=lambda d: d.get("layer_ix"))
    images = []
    for image in list(map(lambda d: d.get("image"), dicts)) + arrays:
        if len(image.shape) != 3:
            image = plt.get_cmap("Greys_r")(image)
        images.append(image)
    fig, ax = plt.subplots()
    ax.imshow(reduce(merge_layers, images), **imshow_kwargs)
    return fig, ax


def simple_figure(
    image: Union[np.ndarray, Image.Image],
    zero_mask: bool = True,
    layers: Optional[Sequence[np.ndarray]] = None,
    **imshow_kwargs,
) -> mpl.figure.Figure:
    """Wrap an array + optional layers up in a simple matplotlib figure."""
    if (zero_mask is True) and isinstance(image, np.ma.MaskedArray):
        image = np.ma.filled(image, 0)
    if layers is not None:
        fig, ax = flatten_into_figure(
            list(layers) + [{"layer_ix": 0, "image": image}], **imshow_kwargs
        )
    else:
        fig, ax = plt.subplots()
        ax.imshow(image, **imshow_kwargs)
    strip_axes(ax)
    return fig


def render_nested_rgb_composite(
    channel_images: Mapping[str, np.ndarray],
    *,
    metadata,
    special_constants=None,
    norm_kwargs=None,
    **channel_instructions: "LookInstruction",
):
    # TODO: is there a cleaner way to handle this import?
    from marslab.imgops.look import Look

    rendered_channels = {}
    for channel, instruction in channel_instructions.items():
        pipeline = Look.compile_from_instruction(
            instruction, metadata, special_constants
        )
        rendered_channels[channel] = pipeline.execute(channel_images[channel])
    norm_kwargs = {} if norm_kwargs is None else norm_kwargs
    norm_channels = [
        normalize_range(rendered_channels[name], **norm_kwargs)
        for name in ("red", "green", "blue")
    ]
    return render_rgb_composite(
        norm_channels, special_constants=special_constants
    )
